#!/usr/bin/env python3
import argparse
import contextlib
import json
import logging
import os
import subprocess
import tempfile
import time
from http.client import HTTPConnection

import requests
import ruamel.yaml as yaml

from paasta_tools.utils import DEFAULT_SOA_CONFIGS_GIT_URL
from paasta_tools.utils import format_git_url
from paasta_tools.utils import load_system_paasta_config

requests_log = logging.getLogger("requests.packages.urllib3")
logging.basicConfig(level=logging.INFO)
log = logging.getLogger()


def parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "-s",
        "--splunk-creds",
        help="Service credentials for Splunk API, user:pass",
        dest="splunk_creds",
        required=True,
    )
    parser.add_argument(
        "-f",
        "--criteria-filter",
        help="Filter Splunk search results criteria field. Default: *",
        dest="criteria_filter",
        required=False,
        default="*",
    )
    parser.add_argument(
        "-j",
        "--jira-creds",
        help="Service credentials for JIRA API, user:pass",
        dest="jira_creds",
        required=False,
    )
    parser.add_argument(
        "-t",
        "--ticket",
        help="Create JIRA tickets for every service in corresponding project (not available in bulk mode)",
        action="store_true",
        dest="ticket",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--reviews",
        help="Guess owners of each service and create reviews automatically",
        action="store_true",
        dest="create_reviews",
        default=False,
    )
    parser.add_argument(
        "-p",
        "--publish-reviews",
        help="Guess owners of each service and publish reviews automatically",
        action="store_true",
        dest="publish_reviews",
        default=False,
    )
    parser.add_argument(
        "-b",
        "--bulk",
        help="Patch all services in the report with only one code review",
        action="store_true",
        dest="bulk",
        default=False,
    )
    parser.add_argument(
        "--app",
        help="Splunk app of the CSV file",
        default="yelp_performance",
        required=False,
        dest="splunk_app",
    )
    parser.add_argument(
        "-c",
        "--csv-report",
        help="Splunk csv file from which to pull data.",
        required=True,
        dest="csv_report",
    )
    parser.add_argument(
        "-y",
        "--yelpsoa-configs-dir",
        help="Use provided existing yelpsoa-configs instead of cloning the repo in a temporary dir. Only avail with -b option",
        dest="YELPSOA_DIR",
        required=False,
    )
    parser.add_argument(
        "-l",
        "--local",
        help="Do not create a branch. Implies -y and -b.",
        action="store_true",
        dest="no_branch",
        default=False,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Debug mode.",
        action="store_true",
        dest="verbose",
    )

    return parser.parse_args()


def tempdir():
    tmp = tempfile.TemporaryDirectory(prefix="repo", dir="/nail/tmp")
    log.debug(f"Created temp directory: {tmp.name}")
    return tmp


@contextlib.contextmanager
def cwd(path):
    pwd = os.getcwd()
    os.chdir(path)
    log.debug(f"Switching from directory {pwd} to {path}")
    try:
        yield
    finally:
        log.debug(f"Switching back from directory {path} to {pwd}")
        os.chdir(pwd)


def _force_str_to_int(value: str) -> int:
    """Force convert strings into ints - even if they're technically floats."""
    return int(float(value))


def get_report_from_splunk(creds, app, filename, criteria_filter):
    """Expect a table containing at least the following fields:
    criteria (<service> kubernetes-<cluster_name> <instance>)
    service_owner (Optional)
    project (Required to create tickets)
    estimated_monthly_savings (Optional)
    search_time (Unix time)
    one of the following pairs:
    - current_cpus and suggested_cpus
    - current_mem and suggested_mem
    - current_disk and suggested_disk
    - suggested_hacheck_cpus
    - suggested_cpu_burst_add
    - suggested_min_instances
    - suggested_max_instances
    """
    url = f"https://splunk-api.yelpcorp.com/servicesNS/nobody/{app}/search/jobs/export"
    search = (
        '| inputlookup {filename} | search criteria="{criteria_filter}"'
        '| eval _time = search_time | where _time > relative_time(now(),"-7d")'
    ).format(filename=filename, criteria_filter=criteria_filter)
    log.debug(f"Sending this query to Splunk: {search}\n")
    data = {"output_mode": "json", "search": search}
    creds = creds.split(":")
    resp = requests.post(url, data=data, auth=(creds[0], creds[1]))
    resp_text = resp.text.split("\n")
    log.info("Found {} services to rightsize".format(len(resp_text) - 1))
    resp_text = [x for x in resp_text if x]
    resp_text = [json.loads(x) for x in resp_text]
    services_to_update = {}

    for d in resp_text:
        if "result" not in d:
            raise ValueError(f"Splunk request didn't return any results: {resp_text}")
        criteria = d["result"]["criteria"]

        serv = {
            "cluster": criteria.split(" ")[1],
            "date": d["result"]["_time"].split(" ")[0],
            "instance": criteria.split(" ")[2],
            "money": d["result"].get("estimated_monthly_savings", 0),
            "old_cpus": d["result"].get("current_cpus"),
            "old_disk": d["result"].get("current_disk"),
            "old_mem": d["result"].get("current_mem"),
            "owner": d["result"].get("service_owner", "Unavailable"),
            "project": d["result"].get("project", "Unavailable"),
            "service": criteria.split(" ")[0],
            # only mergeable fields below
            "cpu_burst_add": d["result"].get("suggested_cpu_burst_add"),
            "cpus": d["result"].get("suggested_cpus"),
            "disk": d["result"].get("suggested_disk"),
            "hacheck_cpus": d["result"].get("suggested_hacheck_cpus"),
            "max_instances": d["result"].get("suggested_max_instances"),
            "mem": d["result"].get("suggested_mem"),
            "min_instances": d["result"].get("suggested_min_instances"),
        }

        # the report we get is all strings, so we need to convert them to the right types
        field_conversions = {
            "cpus": float,
            "cpu_burst_add": float,
            "disk": int,
            "hacheck_cpus": float,
            "max_instances": int,
            "mem": int,
            "min_instances": int,
        }

        # merge results if we've already seen rows for this service
        # NOTE: this is necessary since the Splunk search can return multiple rows
        # for the same (service, cluster, instance) tuple as the autotune query
        # treats certain cpu allocation changes as if the tuple was entirely different.
        # this is ostensibly due to a theory that if you update resource allocation, existing
        # autotune data is potentially invalidated - but in practice this ends up hampering
        # autotune for services with highly variable resource allocation - e.g., we have some services
        # that have their cpu allocation tweaked by +/-.1 cpu pretty frequently, but then min/max autotune
        # is never updated.
        if criteria in services_to_update:
            for key in serv:
                # we probably don't want to merge any other fields since they're going to be strings :p
                if key not in field_conversions:
                    continue

                last_proposed_suggestion = services_to_update[criteria][key]
                proposed_suggestion = serv[key]

                # if both are non-null, take the max of the two
                if (
                    last_proposed_suggestion is not None
                    and proposed_suggestion is not None
                ):
                    services_to_update[criteria][key] = max(
                        last_proposed_suggestion,
                        proposed_suggestion,
                        key=field_conversions[key],
                    )
                # otherwise, if only one of these is non-null, use that one
                elif last_proposed_suggestion is not None:
                    services_to_update[criteria][key] = last_proposed_suggestion
                elif proposed_suggestion is not None:
                    services_to_update[criteria][key] = proposed_suggestion
                # otherwise, if we didn't enter any of the above branches, we're essentially leaving in place the
                # existing None

        # otherwise, simply add the service to the final report
        else:
            services_to_update[criteria] = serv

    return {
        "search": search,
        "results": services_to_update,
    }


def clone_in(target_dir, system_paasta_config=None):
    if not system_paasta_config:
        system_paasta_config = load_system_paasta_config()
    repo_config = system_paasta_config.get_git_repo_config("yelpsoa-configs")

    remote = format_git_url(
        system_paasta_config.get_git_config()["git_user"],
        repo_config.get("git_server", DEFAULT_SOA_CONFIGS_GIT_URL),
        repo_config["repo_name"],
    )
    subprocess.check_call(("git", "clone", remote, target_dir))


def create_branch(branch_name):
    subprocess.check_call(("git", "checkout", "master"))
    subprocess.check_call(("git", "checkout", "-b", branch_name))


def bulk_commit(filenames, originating_search):
    message = f"Rightsizer bulk update\n\nSplunk search:\n{originating_search}"
    subprocess.check_call(["git", "add"] + filenames)
    subprocess.check_call(("git", "commit", "-n", "-m", message))


def bulk_review(filenames, originating_search, publish=False):
    reviewers = set(get_reviewers_in_group("right-sizer"))
    for filename in filenames:
        reviewers = reviewers.union(get_reviewers(filename))

    reviewers_arg = " ".join(list(reviewers))
    summary = "Rightsizer bulk update"
    description = (
        "This is an automated bulk review. It will be shipped automatically if a primary reviewer gives a shipit. If you think this should not be shipped, talk to one of the primary reviewers. \n\n"
        "This review is based on results from the following Splunk search:\n"
        f"{originating_search}"
    )
    review_cmd = [
        "review-branch",
        f"--summary={summary}",
        f"--description={description}",
        "--reviewers",
        reviewers_arg,
        "--server",
        "https://reviewboard.yelpcorp.com",
    ]
    if publish:
        review_cmd.append("-p")

    subprocess.check_call(review_cmd)


def commit(filename, serv):
    message = "Updating {} for {}provisioned cpu from {} to {} cpus".format(
        filename, serv["state"], serv["old_cpus"], serv["cpus"]
    )
    log.debug(f"Commit {filename} with the following message: {message}")
    subprocess.check_call(("git", "add", filename))
    subprocess.check_call(("git", "commit", "-n", "-m", message))


def get_reviewers_in_group(group_name):
    """Using rbt's target-groups argument overrides our configured default review groups.
    So we'll expand the group into usernames and pass those users in the group individually.
    """
    rightsizer_reviewers = json.loads(
        subprocess.check_output(
            (
                "rbt",
                "api-get",
                "--server",
                "https://reviewboard.yelpcorp.com",
                f"groups/{group_name}/users/",
            )
        ).decode("UTF-8")
    )
    return [user.get("username", "") for user in rightsizer_reviewers.get("users", {})]


def get_reviewers(filename):
    recent_authors = set()
    authors = (
        subprocess.check_output(("git", "log", "--format=%ae", "--", filename))
        .decode("UTF-8")
        .splitlines()
    )

    authors = [x.split("@")[0] for x in authors]
    for author in authors:
        if "no-reply" in author:
            continue
        recent_authors.add(author)
        if len(recent_authors) >= 3:
            break
    return recent_authors


def review(filename, summary, description, publish_review):
    all_reviewers = get_reviewers(filename).union(get_reviewers_in_group("right-sizer"))
    reviewers_arg = " ".join(all_reviewers)
    publish_arg = "-p" if publish_review is True else "-d"
    subprocess.check_call(
        (
            "review-branch",
            f"--summary={summary}",
            f"--description={description}",
            publish_arg,
            "--reviewers",
            reviewers_arg,
            "--server",
            "https://reviewboard.yelpcorp.com",
        )
    )


def edit_soa_configs(filename, instance, cpu, mem, disk):
    if os.path.islink(filename):
        real_filename = os.path.realpath(filename)
        os.remove(filename)
    else:
        real_filename = filename
    try:
        with open(real_filename, "r") as fi:
            yams = fi.read()
            yams = yams.replace("cpus: .", "cpus: 0.")
            data = yaml.round_trip_load(yams, preserve_quotes=True)

        instdict = data[instance]
        if cpu:
            instdict["cpus"] = float(cpu)
        if mem:
            mem = max(128, round(float(mem)))
            instdict["mem"] = mem
        if disk:
            instdict["disk"] = round(float(disk))
        out = yaml.round_trip_dump(data, width=120)

        with open(filename, "w") as fi:
            fi.write(out)
    except FileNotFoundError:
        log.exception(f"Could not find {filename}")
    except KeyError:
        log.exception(f"Error in {filename}. Will continue")


def create_jira_ticket(serv, creds, description, JIRA):
    creds = creds.split(":")
    options = {"server": "https://jira.yelpcorp.com"}
    jira_cli = JIRA(options=options, basic_auth=(creds[0], creds[1]))  # noqa: F821
    jira_ticket = {}
    # Sometimes a project has required fields we can't predict
    try:
        jira_ticket = {
            "project": {"key": serv["project"]},
            "description": description,
            "issuetype": {"name": "Improvement"},
            "labels": ["perf-watching", "paasta-rightsizer"],
            "summary": "{s}.{i} in {c} may be {o}provisioned".format(
                s=serv["service"],
                i=serv["instance"],
                c=serv["cluster"],
                o=serv["state"],
            ),
        }
        tick = jira_cli.create_issue(fields=jira_ticket)
    except Exception:
        jira_ticket["project"] = {"key": "PEOBS"}
        jira_ticket["labels"].append(serv["service"])
        tick = jira_cli.create_issue(fields=jira_ticket)
    return tick.key


def _get_dashboard_qs_param(param, value):
    # Some dashboards may ask for query string params like param=value, but not this provider.
    return f"variables%5B%5D={param}%3D{param}:{value}"


def generate_ticket_content(serv):
    cpus = float(serv["cpus"])
    provisioned_state = "over"
    if cpus > float(serv["old_cpus"]):
        provisioned_state = "under"

    serv["state"] = provisioned_state
    ticket_desc = (
        "This ticket and CR have been auto-generated to help keep PaaSTA right-sized."
        "\nPEOBS will review this CR and give a shipit. Then an ops deputy from your team can merge"
        " if these values look good for your service after review."
        "\nOpen an issue with any concerns and someone from PEOBS will respond."
        "\nWe suspect that {s}.{i} in {c} may have been {o}-provisioned"
        " during the 1 week prior to {d}. It initially had {x} cpus, but based on the below dashboard,"
        " we recommend {y} cpus."
        "\n- Dashboard: https://y.yelpcorp.com/{o}provisioned?{cluster_param}&{service_param}&{instance_param}"
        "\n- Service owner: {n}"
        "\n- Estimated monthly excess cost: ${m}"
        "\n\nFor more information and sizing examples for larger services:"
        "\n- Runbook: https://y.yelpcorp.com/rb-provisioning-alert"
        "\n- Alert owner: pe-observability@yelp.com"
    ).format(
        s=serv["service"],
        c=serv["cluster"],
        i=serv["instance"],
        o=provisioned_state,
        d=serv["date"],
        n=serv["owner"],
        m=serv["money"],
        x=serv["old_cpus"],
        y=serv["cpus"],
        cluster_param=_get_dashboard_qs_param("paasta_cluster", serv["cluster"]),
        service_param=_get_dashboard_qs_param("paasta_service", serv["service"]),
        instance_param=_get_dashboard_qs_param("paasta_instance", serv["instance"]),
    )
    summary = f"Rightsizing {serv['service']}.{serv['instance']} in {serv['cluster']} to make it not have {provisioned_state}-provisioned cpu"  # noqa: E501
    return (summary, ticket_desc)


def bulk_rightsize(report, create_code_review, publish_code_review, create_new_branch):
    if create_new_branch:
        branch = "rightsize-bulk-{}".format(int(time.time()))
        create_branch(branch)

    filenames = []
    for _, serv in report["results"].items():
        filename = "{}/{}.yaml".format(serv["service"], serv["cluster"])
        filenames.append(filename)
        cpus = serv.get("cpus", None)
        mem = serv.get("mem", None)
        disk = serv.get("disk", None)
        edit_soa_configs(filename, serv["instance"], cpus, mem, disk)
    if create_code_review:
        bulk_commit(filenames, report["search"])
        bulk_review(filenames, report["search"], publish_code_review)


def individual_rightsize(
    report, create_tickets, jira_creds, create_review, publish_review, JIRA
):
    for _, serv in report["results"].items():
        filename = "{}/{}.yaml".format(serv["service"], serv["cluster"])
        summary, ticket_desc = generate_ticket_content(serv)

        if create_tickets is True:
            branch = create_jira_ticket(serv, jira_creds, ticket_desc, JIRA)
        else:
            branch = "rightsize-{}".format(int(time.time() * 1000))

        create_branch(branch)
        cpus = serv.get("cpus", None)
        mem = serv.get("mem", None)
        disk = serv.get("disk", None)
        edit_soa_configs(filename, serv["instance"], cpus, mem, disk)
        try:
            commit(filename, serv)
            if create_review:
                review(filename, summary, ticket_desc, publish_review)
        except Exception:
            log.exception(
                (
                    "\nUnable to push changes to {f}. Check if {f} conforms to"
                    "yelpsoa-configs yaml rules. No review created. To see the"
                    "cpu suggestion for this service check {t}."
                ).format(f=filename, t=branch)
            )
            continue


def main():
    args = parse_args()

    if args.verbose:
        log.setLevel(logging.DEBUG)
        requests_log.setLevel(logging.DEBUG)
        HTTPConnection.debuglevel = 2
        requests_log.propagate = True

    # Safety checks
    if args.no_branch and not args.YELPSOA_DIR:
        log.error(
            "You must specify --yelpsoa-configs-dir to work on if you use the --local option"
        )
        return False

    if args.ticket:
        if not args.jira_creds:
            raise ValueError("No JIRA creds specified")
        # Only import the jira module if we need too
        from jira.client import JIRA  # noqa: F401
    else:
        JIRA = None

    report = get_report_from_splunk(
        args.splunk_creds, args.splunk_app, args.csv_report, args.criteria_filter
    )

    tmpdir = tempdir()  # Create a tmp dir even if we are not using it

    working_dir = args.YELPSOA_DIR
    system_paasta_config = load_system_paasta_config()
    if working_dir is None:
        # Working in a temporary directory
        working_dir = os.path.join("rightsizer", tmpdir.name)
        clone_in(working_dir, system_paasta_config=system_paasta_config)

    with cwd(working_dir):
        if args.bulk or args.no_branch:
            log.info("Running in bulk mode")
            bulk_rightsize(
                report, args.create_reviews, args.publish_reviews, not args.no_branch
            )
        else:
            individual_rightsize(
                report,
                args.ticket,
                args.jira_creds,
                args.create_reviews,
                args.publish_reviews,
                JIRA,
            )

    tmpdir.cleanup()  # Cleanup any tmpdire used


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
from unittest import mock
from unittest.mock import AsyncMock
from unittest.mock import MagicMock

import pytest

from paasta_tools import drain_lib


def test_register_drain_method():

    with mock.patch.dict(drain_lib._drain_methods):

        @drain_lib.register_drain_method("FAKEDRAINMETHOD")
        class FakeDrainMethod(drain_lib.DrainMethod):
            pass

        assert (
            type(drain_lib.get_drain_method("FAKEDRAINMETHOD", "srv", "inst", "ns"))
            == FakeDrainMethod
        )


@contextlib.contextmanager
def mock_ClientSession(**fake_session_kwargs):
    fake_session = MagicMock(name="session", **fake_session_kwargs)

    class FakeClientSession:
        def __init__(self, *args, **kwargs):
            ...

        async def __aenter__(*args):
            return fake_session

        async def __aexit__(*args):
            pass

    with mock.patch("aiohttp.ClientSession", new=FakeClientSession, autospec=False):
        yield


class TestHacheckDrainMethod:
    drain_method = drain_lib.HacheckDrainMethod(
        service="srv",
        instance="inst",
        registrations=["ns_one", "ns_two"],
        hacheck_port=12345,
    )

    async def _async_id(self, x):
        return x

    def test_spool_urls(self):
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        actual = self.drain_method.spool_urls(fake_task)
        # Nerve hits /{mode}/{service}.{namespace}/{port}/status
        expected = [
            f"http://fake_host:12345/spool/{ns}/54321/status"
            for ns in self.drain_method.registrations
        ]
        assert actual == expected

    @pytest.mark.asyncio
    async def test_for_each_registration_with_no_ports(self):
        fake_task = mock.Mock(host="fake_host", ports=[])
        actual = await self.drain_method.for_each_registration(
            task=fake_task, func=self._async_id
        )
        assert actual is None

    @pytest.mark.asyncio
    async def test_for_each_registration(self):
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        actual = await self.drain_method.for_each_registration(
            task=fake_task, func=self._async_id
        )
        assert actual == self.drain_method.spool_urls(fake_task)

    @pytest.mark.asyncio
    async def test_is_draining_yes(self):
        fake_response = mock.Mock(
            status=503,
            text=AsyncMock(
                return_value="Service service in down state since 1435694078.778886 "
                "until 1435694178.780000: Drained by Paasta"
            ),
        )
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        with mock_ClientSession(
            get=mock.Mock(
                return_value=MagicMock(__aenter__=AsyncMock(return_value=fake_response))
            )
        ):
            assert await self.drain_method.is_draining(fake_task) is True

    @pytest.mark.asyncio
    async def test_is_draining_no(self):
        fake_response = mock.Mock(status=200, text=AsyncMock(return_value=""))
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        with mock_ClientSession(
            get=mock.Mock(
                return_value=MagicMock(__aenter__=AsyncMock(return_value=fake_response))
            )
        ):
            assert await self.drain_method.is_draining(fake_task) is False


class TestHTTPDrainMethod:
    def test_get_format_params(self):
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        drain_method = drain_lib.HTTPDrainMethod(
            "fake_service", "fake_instance", ["fake_nerve_ns"], {}, {}, {}, {}
        )
        assert drain_method.get_format_params(fake_task) == [
            {
                "host": "fake_host",
                "port": 54321,
                "service": "fake_service",
                "instance": "fake_instance",
                "nerve_ns": "fake_nerve_ns",
            }
        ]

    def test_format_url(self):
        drain_method = drain_lib.HTTPDrainMethod(
            "fake_service", "fake_instance", ["fake_nerve_ns"], {}, {}, {}, {}
        )
        url_format = "foo_{host}"
        format_params = {"host": "fake_host"}
        assert drain_method.format_url(url_format, format_params) == "foo_fake_host"

    def test_parse_success_codes(self):
        drain_method = drain_lib.HTTPDrainMethod(
            "fake_service", "fake_instance", ["fake_nerve_ns"], {}, {}, {}, {}
        )
        assert drain_method.parse_success_codes("200") == {200}
        assert drain_method.parse_success_codes("200-203") == {200, 201, 202, 203}
        assert drain_method.parse_success_codes("200-202,302,305-306") == {
            200,
            201,
            202,
            302,
            305,
            305,
            306,
        }
        assert drain_method.parse_success_codes(200) == {200}

    def test_check_response_code(self):
        drain_method = drain_lib.HTTPDrainMethod(
            "fake_service", "fake_instance", ["fake_nerve_ns"], {}, {}, {}, {}
        )

        # Happy case
        assert drain_method.check_response_code(200, "200-299") is True

        # Sad case
        assert drain_method.check_response_code(500, "200-299") is False

    @pytest.mark.asyncio
    async def test_issue_request(self):
        drain_method = drain_lib.HTTPDrainMethod(
            "fake_service", "fake_instance", ["fake_nerve_ns"], {}, {}, {}, {}
        )
        fake_task = mock.Mock(host="fake_host", ports=[54321])
        url_spec = {
            "url_format": "http://localhost:654321/fake/{host}",
            "method": "get",
            "success_codes": "1234",
        }

        fake_resp = mock.Mock(status=1234)
        mock_request = mock.Mock(return_value=AsyncMock(return_value=fake_resp)())
        with mock_ClientSession(request=mock_request):
            await drain_method.issue_request(url_spec=url_spec, task=fake_task)

        mock_request.assert_called_once_with(
            method="GET",
            url="http://localhost:654321/fake/fake_host",
            headers=mock.ANY,
            timeout=15,
        )



from unittest import mock

import pytest

from paasta_tools.autoscaling.utils import MetricsProviderDict
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_ACTIVE_REQUESTS
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_CPU
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_GUNICORN
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_UWSGI
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_UWSGI_V2
from paasta_tools.long_running_service_tools import METRICS_PROVIDER_WORKER_LOAD
from paasta_tools.setup_prometheus_adapter_config import _minify_promql
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_active_requests_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_arbitrary_promql_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_gunicorn_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_uwsgi_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_uwsgi_v2_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import (
    create_instance_worker_load_scaling_rule,
)
from paasta_tools.setup_prometheus_adapter_config import get_rules_for_service_instance
from paasta_tools.utils import SystemPaastaConfig

MOCK_SYSTEM_PAASTA_CONFIG = SystemPaastaConfig(
    {},
    "/mock/system/configs",
)


@pytest.mark.parametrize(
    "registrations,expected_instance",
    [
        (
            ["test_service.abc", "test_service.xyz", "test_service.123"],
            "test_instance",
        ),
        (
            ["test_service.xyz"],
            "xyz",
        ),
    ],
)
def test_create_instance_active_requests_scaling_rule(
    registrations: list, expected_instance: str
) -> None:
    service_name = "test_service"
    instance_config = mock.Mock(
        instance="test_instance",
        get_registrations=mock.Mock(return_value=registrations),
    )
    metrics_provider_config = MetricsProviderDict(
        {
            "type": METRICS_PROVIDER_ACTIVE_REQUESTS,
            "desired_active_requests_per_replica": 12,
            "moving_average_window_seconds": 20120302,
        }
    )
    paasta_cluster = "test_cluster"
    rule = create_instance_active_requests_scaling_rule(
        service=service_name,
        instance_config=instance_config,
        metrics_provider_config=metrics_provider_config,
        paasta_cluster=paasta_cluster,
    )

    # we test that the format of the dictionary is as expected with mypy
    # and we don't want to test the full contents of the retval since then
    # we're basically just writting a change-detector test - instead, we test
    # that we're actually using our inputs
    assert service_name in rule["seriesQuery"]
    assert instance_config.instance in rule["seriesQuery"]
    assert paasta_cluster in rule["seriesQuery"]
    # these two numbers are distinctive and unlikely to be used as constants
    assert (
        str(metrics_provider_config["desired_active_requests_per_replica"])
        in rule["metricsQuery"]
    )
    assert (
        str(metrics_provider_config["moving_average_window_seconds"])
        in rule["metricsQuery"]
    )
    assert f"paasta_instance='{expected_instance}'" in rule["metricsQuery"]


def test_create_instance_uwsgi_scaling_rule() -> None:
    service_name = "test_service"
    instance_config = mock.Mock(instance="test_instance")
    metrics_provider_config = MetricsProviderDict(
        {
            "type": METRICS_PROVIDER_UWSGI,
            "setpoint": 0.1234567890,
            "moving_average_window_seconds": 20120302,
        }
    )
    paasta_cluster = "test_cluster"
    rule = create_instance_uwsgi_scaling_rule(
        service=service_name,
        instance_config=instance_config,
        metrics_provider_config=metrics_provider_config,
        paasta_cluster=paasta_cluster,
    )

    # we test that the format of the dictionary is as expected with mypy
    # and we don't want to test the full contents of the retval since then
    # we're basically just writting a change-detector test - instead, we test
    # that we're actually using our inputs
    assert service_name in rule["seriesQuery"]
    assert instance_config.instance in rule["seriesQuery"]
    assert paasta_cluster in rule["seriesQuery"]
    # these two numbers are distinctive and unlikely to be used as constants
    assert str(metrics_provider_config["setpoint"]) in rule["metricsQuery"]
    assert (
        str(metrics_provider_config["moving_average_window_seconds"])
        in rule["metricsQuery"]
    )


def test_create_instance_uwsgi_v2_scaling_rule() -> None:
    service_name = "test_service"
    instance_config = mock.Mock(instance="test_instance")
    metrics_provider_config = MetricsProviderDict(
        {
            "type": METRICS_PROVIDER_UWSGI_V2,
            "setpoint": 0.1234567890,
            "moving_average_window_seconds": 20120302,
        }
    )
    paasta_cluster = "test_cluster"
    rule = create_instance_uwsgi_v2_scaling_rule(
        service=service_name,
        instance_config=instance_config,
        metrics_provider_config=metrics_provider_config,
        paasta_cluster=paasta_cluster,
    )

    # we test that the format of the dictionary is as expected with mypy
    # and we don't want to test the full contents of the retval since then
    # we're basically just writing a change-detector test - instead, we test
    # that we're actually using our inputs
    assert service_name in rule["seriesQuery"]
    assert instance_config.instance in rule["seriesQuery"]
    assert paasta_cluster in rule["seriesQuery"]

    # Unlike uwsgi(v1), we don't use the setpoint in this query -- the HPA will have the setpoint as its target.
    assert str(metrics_provider_config["setpoint"]) not in rule["metricsQuery"]
    assert (
        str(metrics_provider_config["moving_average_window_seconds"])
        in rule["metricsQuery"]
    )


def test_create_instance_worker_load_scaling_rule() -> None:
    service_name = "test_service"
    instance_config = mock.Mock(instance="test_instance")
    metrics_provider_config = MetricsProviderDict(
        {
            "type": METRICS_PROVIDER_WORKER_LOAD,
            "setpoint": 0.1234567890,
            "moving_average_window_seconds": 20120302,
        }
    )
    paasta_cluster = "test_cluster"
    rule = create_instance_worker_load_scaling_rule(
        service=service_name,
        instance_config=instance_config,
        metrics_provider_config=metrics_provider_config,
        paasta_cluster=paasta_cluster,
    )

    # we test that the format of the dictionary is as expected with mypy
    # and we don't want to test the full contents of the retval since then
    # we're basically just writing a change-detector test - instead, we test
    # that we're actually using our inputs
    assert service_name in rule["seriesQuery"]
    assert instance_config.instance in rule["seriesQuery"]
    assert paasta_cluster in rule["seriesQuery"]

    # Like uwsgi_v2, we don't use the setpoint in this query -- the HPA will have the setpoint as its target.
    assert str(metrics_provider_config["setpoint"]) not in rule["metricsQuery"]
    assert (
        str(metrics_provider_config["moving_average_window_seconds"])
        in rule["metricsQuery"]
    )

    # Verify that we're using the generic worker_busy metric instead of uwsgi_worker_busy
    assert "worker_busy" in rule["seriesQuery"]
    assert "worker_busy" in rule["metricsQuery"]
    assert "uwsgi_worker_busy" not in rule["seriesQuery"]
    assert "uwsgi_worker_busy" not in rule["metricsQuery"]

    # Verify the metric name uses worker-load-prom suffix
    assert rule["name"]["as"].endswith("-worker-load-prom")


def test_create_instance_gunicorn_scaling_rule() -> None:
    service_name = "test_service"
    instance_config = mock.Mock(instance="test_instance")
    metrics_provider_config = MetricsProviderDict(
        {
            "type": METRICS_PROVIDER_GUNICORN,
            "setpoint": 0.1234567890,
            "moving_average_window_seconds": 20120302,
        }
    )
    paasta_cluster = "test_cluster"
    rule = create_instance_gunicorn_scaling_rule(
        service=service_name,
        instance_config=instance_config,
        metrics_provider_config=metrics_provider_config,
        paasta_cluster=paasta_cluster,
    )

    # we test that the format of the dictionary is as expected with mypy
    # and we don't want to test the full contents of the retval since then
    # we're basically just writting a change-detector test - instead, we test
    # that we're actually using our inputs
    assert service_name in rule["seriesQuery"]
    assert instance_config.instance in rule["seriesQuery"]
    assert paasta_cluster in rule["seriesQuery"]
    # these two numbers are distinctive and unlikely to be used as constants
    assert str(metrics_provider_config["setpoint"]) in rule["metricsQuery"]
    assert (
        str(metrics_provider_config["moving_average_window_seconds"])
        in rule["metricsQuery"]
    )


@pytest.mark.parametrize(
    "instance_config,expected_rules",
    [
        (
            mock.Mock(
                instance="instance",
                get_namespace=mock.Mock(return_value="test_namespace"),
                get_autoscaling_metrics_provider=mock.Mock(
                    side_effect=lambda x: (
                        {
                            "type": METRICS_PROVIDER_CPU,
                            "setpoint": 0.1234567890,
                            "moving_average_window_seconds": 20120302,
                        }
                        if x == METRICS_PROVIDER_CPU
                        else None
                    )
                ),
            ),
            0,
        ),
        (
            mock.Mock(
                instance="instance",
                get_namespace=mock.Mock(return_value="test_namespace"),
                get_autoscaling_metrics_provider=mock.Mock(
                    side_effect=lambda x: (
                        {
                            "type": METRICS_PROVIDER_UWSGI,
                            "setpoint": 0.1234567890,
                            "moving_average_window_seconds": 20120302,
                        }
                        if x == METRICS_PROVIDER_UWSGI
                        else None
                    )
                ),
            ),
            1,
        ),
        (
            mock.Mock(
                instance="instance",
                get_namespace=mock.Mock(return_value="test_namespace"),
                get_autoscaling_metrics_provider=mock.Mock(
                    side_effect=lambda x: (
                        {
                            "type": METRICS_PROVIDER_UWSGI,
                            "setpoint": 0.1234567890,
                            "moving_average_window_seconds": 20120302,
                        }
                        if x == METRICS_PROVIDER_UWSGI
                        else (
                            {
                                "type": METRICS_PROVIDER_CPU,
                                "setpoint": 0.1234567890,
                                "moving_average_window_seconds": 20120302,
                            }
                            if x == METRICS_PROVIDER_CPU
                            else None
                        )
                    )
                ),
            ),
            1,
        ),
        (
            mock.Mock(
                instance="instance",
                get_namespace=mock.Mock(return_value="test_namespace"),
                get_autoscaling_metrics_provider=mock.Mock(
                    side_effect=lambda x: (
                        {
                            "type": METRICS_PROVIDER_UWSGI,
                            "setpoint": 0.1234567890,
                            "moving_average_window_seconds": 20120302,
                        }
                        if x == METRICS_PROVIDER_UWSGI
                        else (
                            {
                                "type": METRICS_PROVIDER_GUNICORN,
                                "setpoint": 0.1234567890,
                                "moving_average_window_seconds": 20120302,
                            }
                            if x == METRICS_PROVIDER_GUNICORN
                            else None
                        )
                    )
                ),
            ),
            2,
        ),
        (
            mock.Mock(
                instance="instance",
                get_namespace=mock.Mock(return_value="test_namespace"),
                get_autoscaling_metrics_provider=mock.Mock(
                    side_effect=lambda x: (
                        {
                            "type": METRICS_PROVIDER_WORKER_LOAD,
                            "setpoint": 0.1234567890,
                            "moving_average_window_seconds": 20120302,
                        }
                        if x == METRICS_PROVIDER_WORKER_LOAD
                        else None
                    )
                ),
            ),
            1,
        ),
    ],
)
def test_get_rules_for_service_instance(
    instance_config: KubernetesDeploymentConfig,
    expected_rules: int,
) -> None:
    assert (
        len(
            get_rules_for_service_instance(
                service_name="service",
                instance_config=instance_config,
                paasta_cluster="cluster",
            )
        )
        == expected_rules
    )


@pytest.mark.parametrize(
    "query,expected",
    [
        # empty strings shouldn't be touched (or happen...)
        ("", ""),
        # we should have no leading/trailing whitespace
        ("\t\t\ttest{label='a'}\n", "test{label='a'}"),
        # we should collapse internal whitespace
        ("\t \ttest{label='a'}\n   test{label='b'}", "test{label='a'} test{label='b'}"),
        # we shouldn't touch whitespace inside of labels
        ("\t \ttest{label='a  b'}\n", "test{label='a  b'}"),
    ],
)
def test__minify_promql(query: str, expected: str) -> None:
    assert _minify_promql(query) == expected


def test_create_instance_arbitrary_promql_scaling_rule_no_seriesQuery():
    rule = create_instance_arbitrary_promql_scaling_rule(
        service="service",
        instance_config=mock.Mock(
            instance="instance",
            get_namespace=mock.Mock(return_value="paasta"),
        ),
        metrics_provider_config={"prometheus_adapter_config": {"metricsQuery": "foo"}},
        paasta_cluster="cluster",
    )

    assert rule == {
        "name": {"as": "service-instance-arbitrary-promql"},
        "resources": {
            "overrides": {
                "namespace": {"resource": "namespace"},
                "deployment": {"group": "apps", "resource": "deployments"},
            },
        },
        "metricsQuery": "label_replace( label_replace( foo, 'deployment', 'service-instance', '', '' ), 'namespace', 'paasta', '', '' )",
        "seriesQuery": "kube_deployment_labels{ deployment='service-instance', paasta_cluster='cluster', namespace='paasta' }",
    }


def test_create_instance_arbitrary_promql_scaling_rule_with_seriesQuery():
    rule = create_instance_arbitrary_promql_scaling_rule(
        service="service",
        instance_config=mock.Mock(
            instance="instance",
            get_namespace=mock.Mock(return_value="test_namespace"),
        ),
        metrics_provider_config={
            "prometheus_adapter_config": {
                "metricsQuery": "foo",
                "seriesQuery": "bar",
            }
        },
        paasta_cluster="cluster",
    )

    assert rule == {
        "name": {"as": "service-instance-arbitrary-promql"},
        "resources": {
            "overrides": {
                "namespace": {"resource": "namespace"},
                "deployment": {"group": "apps", "resource": "deployments"},
            },
        },
        "metricsQuery": "foo",  # if seriesQuery is specified, the user's metricsQuery should be unaltered.
        "seriesQuery": "bar",
    }



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
from typing import NamedTuple
from typing import Optional

import cachetools.func
import pyramid
import requests
from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPForbidden
from pyramid.registry import Registry
from pyramid.request import Request
from pyramid.response import Response

from paasta_tools.api.tweens import Handler


logger = logging.getLogger(__name__)
AUTH_CACHE_SIZE = 50000
AUTH_CACHE_TTL = 30 * 60


class AuthorizationOutcome(NamedTuple):
    authorized: bool
    reason: str


class AuthTweenFactory:
    def __init__(self, handler: Handler, registry: Registry) -> None:
        self.handler = handler
        self.registry = registry
        self.enforce = bool(os.getenv("PAASTA_API_AUTH_ENFORCE", ""))
        self.endpoint = os.getenv("PAASTA_API_AUTH_ENDPOINT")
        self.session = requests.Session()

    def __call__(self, request: Request) -> Response:
        """
        Extracts relevant metadata from request, and checks if it is authorized
        """
        token = request.headers.get("Authorization", "").strip()
        token = token.split()[-1] if token else ""  # removes "Bearer" prefix
        auth_outcome = self.is_request_authorized(
            request.path,
            token,
            request.method,
            request.swagger_data.get("service", None),
        )
        if self.enforce and not auth_outcome.authorized:
            return HTTPForbidden(
                body=json.dumps({"reason": auth_outcome.reason}),
                headers={"X-Auth-Failure-Reason": auth_outcome.reason},
                content_type="application/json",
                charset="utf-8",
            )
        return self.handler(request)

    @cachetools.func.ttl_cache(maxsize=AUTH_CACHE_SIZE, ttl=AUTH_CACHE_TTL)
    def is_request_authorized(
        self,
        path: str,
        token: str,
        method: str,
        service: Optional[str],
    ) -> AuthorizationOutcome:
        """
        Check if API request is authorized

        :param str path: API path
        :param str token: authentication token
        :param str method: http method
        :return: auth outcome
        """
        try:
            response = self.session.post(
                url=self.endpoint,
                json={
                    "input": {
                        "path": path,
                        "backend": "paasta",
                        "token": token,
                        "method": method,
                        "service": service,
                    },
                },
                timeout=2,
            ).json()
        except Exception as e:
            logger.exception(f"Issue communicating with auth endpoint: {e}")
            return AuthorizationOutcome(False, "Auth backend error")

        auth_result_allowed = response.get("result", {}).get("allowed")
        if auth_result_allowed is None:
            return AuthorizationOutcome(False, "Malformed auth response")

        if not auth_result_allowed:
            reason = response["result"].get("reason", "Denied")
            return AuthorizationOutcome(False, reason)

        reason = response["result"].get("reason", "Ok")
        return AuthorizationOutcome(True, reason)


def includeme(config: Configurator):
    if os.getenv("PAASTA_API_AUTH_ENDPOINT"):
        config.add_tween(
            "paasta_tools.api.tweens.auth.AuthTweenFactory",
            under=(
                pyramid.tweens.INGRESS,
                "pyramid_swagger.tween.validation_tween_factory",
            ),
        )



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import pysensu_yelp
import pytest

from paasta_tools import check_flink_services_health
from paasta_tools import check_services_replication_tools
from paasta_tools.check_flink_services_health import check_under_registered_taskmanagers
from paasta_tools.utils import compose_job_id

check_flink_services_health.log = mock.Mock()
check_services_replication_tools.log = mock.Mock()


@pytest.fixture
def instance_config():
    service = "fake_service"
    instance = "fake_instance"
    job_id = compose_job_id(service, instance)
    mock_instance_config = mock.Mock(
        service=service,
        instance=instance,
        cluster="fake_cluster",
        soa_dir="fake_soa_dir",
        job_id=job_id,
        config_dict={},
    )
    mock_instance_config.get_replication_crit_percentage.return_value = 100
    mock_instance_config.get_registrations.return_value = [job_id]
    return mock_instance_config


@mock.patch(
    "paasta_tools.flink_tools.get_flink_jobmanager_overview",
    autospec=True,
    return_value={"taskmanagers": 3},
)
def test_check_under_registered_taskmanagers_ok(mock_overview, instance_config):
    under, output, description = check_under_registered_taskmanagers(
        instance_config,
        expected_count=3,
        cr_name="fake--service-575c857546",
        is_eks=False,
    )
    assert not under
    assert (
        "fake_service.fake_instance has 3/3 taskmanagers "
        "reported by dashboard (threshold: 100%)"
    ) in output
    assert "fake_service.fake_instance taskmanager is available" in description


@mock.patch(
    "paasta_tools.flink_tools.get_flink_jobmanager_overview",
    autospec=True,
    return_value={"taskmanagers": 2},
)
def test_check_under_registered_taskmanagers_under(mock_overview, instance_config):
    under, output, description = check_under_registered_taskmanagers(
        instance_config,
        expected_count=3,
        cr_name="fake--service-575c857546",
        is_eks=False,
    )
    assert under
    assert (
        "fake_service.fake_instance has 2/3 taskmanagers "
        "reported by dashboard (threshold: 100%)"
    ) in output
    assert (
        "paasta status -s fake_service -i fake_instance -c fake_cluster -vv"
        in description
    )


@mock.patch(
    "paasta_tools.flink_tools.get_flink_jobmanager_overview",
    autospec=True,
    side_effect=ValueError("dummy exception"),
)
def test_check_under_registered_taskmanagers_error(mock_overview, instance_config):
    under, output, description = check_under_registered_taskmanagers(
        instance_config,
        expected_count=3,
        cr_name="fake--service-575c857546",
        is_eks=False,
    )
    assert under
    assert (
        "Dashboard of service fake_service.fake_instance is not available "
        "(dummy exception)"
    ) in output
    assert (
        "paasta status -s fake_service -i fake_instance -c fake_cluster -vv"
        in description
    )


def test_check_flink_service_health_healthy(instance_config):
    with mock.patch(
        "paasta_tools.check_flink_services_health.healthy_flink_containers_cnt",
        autospec=True,
        return_value=1,
    ), mock.patch(
        "paasta_tools.check_flink_services_health.check_under_replication",
        autospec=True,
        return_value=(False, "OK", "check_rep"),
    ) as mock_check_under_replication, mock.patch(
        "paasta_tools.check_flink_services_health.check_under_registered_taskmanagers",
        autospec=True,
        return_value=(False, "OK", "check_task"),
    ) as mock_check_under_registered_taskmanagers, mock.patch(
        "paasta_tools.check_flink_services_health.send_replication_event", autospec=True
    ) as mock_send_replication_event:
        instance_config.config_dict["taskmanager"] = {"instances": 3}
        check_flink_services_health.check_flink_service_health(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        expected = [
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="supervisor",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="jobmanager",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=3,
                num_available=1,
                sub_component="taskmanager",
            ),
        ]
        mock_check_under_replication.assert_has_calls(expected)
        mock_check_under_registered_taskmanagers.assert_called_once_with(
            instance_config=instance_config, expected_count=3, cr_name="", is_eks=False
        )
        mock_send_replication_event.assert_called_once_with(
            instance_config=instance_config,
            status=pysensu_yelp.Status.OK,
            output="OK, OK, OK, OK",
            description="check_rep\n########\ncheck_rep\n########\ncheck_rep\n########\ncheck_task",
            dry_run=True,
        )


def test_check_flink_service_health_too_few_taskmanagers(instance_config):
    def check_under_replication_side_effect(*args, **kwargs):
        if kwargs["sub_component"] == "supervisor":
            return False, "OK", "check_rep"
        if kwargs["sub_component"] == "jobmanager":
            return False, "OK", "check_rep"
        if kwargs["sub_component"] == "taskmanager":
            return True, "NOPE", "check_rep"

    with mock.patch(
        "paasta_tools.check_flink_services_health.healthy_flink_containers_cnt",
        autospec=True,
        return_value=1,
    ), mock.patch(
        "paasta_tools.check_flink_services_health.check_under_registered_taskmanagers",
        autospec=True,
        return_value=(True, "NOPE", "check_task"),
    ) as mock_check_under_registered_taskmanagers, mock.patch(
        "paasta_tools.check_flink_services_health.check_under_replication",
        autospec=True,
        side_effect=check_under_replication_side_effect,
    ) as mock_check_under_replication, mock.patch(
        "paasta_tools.check_flink_services_health.send_replication_event", autospec=True
    ) as mock_send_replication_event:
        instance_config.config_dict["taskmanager"] = {"instances": 3}
        check_flink_services_health.check_flink_service_health(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        expected = [
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="supervisor",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="jobmanager",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=3,
                num_available=1,
                sub_component="taskmanager",
            ),
        ]
        mock_check_under_replication.assert_has_calls(expected)
        mock_check_under_registered_taskmanagers.assert_called_once_with(
            instance_config=instance_config, expected_count=3, cr_name="", is_eks=False
        )
        mock_send_replication_event.assert_called_once_with(
            instance_config=instance_config,
            status=pysensu_yelp.Status.CRITICAL,
            output="OK, OK, NOPE, NOPE",
            description="check_rep\n########\ncheck_rep\n########\ncheck_rep\n########\ncheck_task",
            dry_run=True,
        )


def test_check_flink_service_health_under_registered_taskamanagers(instance_config):
    with mock.patch(
        "paasta_tools.check_flink_services_health.healthy_flink_containers_cnt",
        autospec=True,
        return_value=1,
    ), mock.patch(
        "paasta_tools.check_flink_services_health.check_under_replication",
        autospec=True,
        return_value=(False, "OK", "check_rep"),
    ) as mock_check_under_replication, mock.patch(
        "paasta_tools.check_flink_services_health.check_under_registered_taskmanagers",
        autospec=True,
        return_value=(True, "NOPE", "check_task"),
    ) as mock_check_under_registered_taskmanagers, mock.patch(
        "paasta_tools.check_flink_services_health.send_replication_event", autospec=True
    ) as mock_send_replication_event:
        instance_config.config_dict["taskmanager"] = {"instances": 3}
        check_flink_services_health.check_flink_service_health(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        expected = [
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="supervisor",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=1,
                num_available=1,
                sub_component="jobmanager",
            ),
            mock.call(
                instance_config=instance_config,
                expected_count=3,
                num_available=1,
                sub_component="taskmanager",
            ),
        ]
        mock_check_under_replication.assert_has_calls(expected)
        mock_check_under_registered_taskmanagers.assert_called_once_with(
            instance_config=instance_config,
            expected_count=3,
            cr_name="",
            is_eks=False,
        )
        mock_send_replication_event.assert_called_once_with(
            instance_config=instance_config,
            status=pysensu_yelp.Status.CRITICAL,
            output="OK, OK, OK, NOPE",
            description="check_rep\n########\ncheck_rep\n########\ncheck_rep\n########\ncheck_task",
            dry_run=True,
        )



#!/opt/venvs/paasta-tools/bin/python
import argparse
import json
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import a_sync
from kubernetes.client import V1Pod
from kubernetes.client import V1ResourceRequirements

from paasta_tools import kubernetes_tools
from paasta_tools import mesos_tools
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.mesos.exceptions import SlaveDoesNotExist
from paasta_tools.mesos.task import Task
from paasta_tools.utils import load_system_paasta_config


MAIN_CONTAINER_TYPE = "main"


class TaskAllocationInfo(NamedTuple):
    paasta_service: str
    paasta_instance: str
    container_type: str
    paasta_pool: str
    resources: Mapping[str, float]
    start_time: float
    docker_id: str
    pod_name: str
    pod_ip: str
    host_ip: str
    git_sha: str
    config_sha: str
    mesos_container_id: str  # Because Mesos task info does not have docker id
    namespace: Optional[str]


def get_container_info_from_mesos_task(
    task: Task,
) -> Tuple[Optional[str], Optional[float]]:
    for status in task["statuses"]:
        if status["state"] != "TASK_RUNNING":
            continue
        container_id = (
            status.get("container_status", {}).get("container_id", {}).get("value")
        )
        time_start = status.get("timestamp")
        return container_id, time_start
    return None, None


def get_paasta_service_instance_from_mesos_task(
    task: Task,
) -> Tuple[Optional[str], Optional[str]]:
    try:
        docker_params = task["container"].get("docker", {}).get("parameters", [])
    except KeyError:
        return None, None
    service, instance = None, None
    for param in docker_params:
        if param["key"] == "label":
            label = param["value"]
            if label.startswith("paasta_service="):
                service = label.split("=")[1]
            if label.startswith("paasta_instance="):
                instance = label.split("=")[1]
    return service, instance


async def get_pool_from_mesos_task(task: Task) -> Optional[str]:
    try:
        attributes = (await task.slave())["attributes"]
        return attributes.get("pool", "default")
    except SlaveDoesNotExist:
        return None


@a_sync.to_blocking
async def get_mesos_task_allocation_info() -> Iterable[TaskAllocationInfo]:
    tasks = await mesos_tools.get_cached_list_of_running_tasks_from_frameworks()
    info_list = []
    for task in tasks:
        mesos_container_id, start_time = get_container_info_from_mesos_task(task)
        paasta_service, paasta_instance = get_paasta_service_instance_from_mesos_task(
            task
        )
        paasta_pool = await get_pool_from_mesos_task(task)
        info_list.append(
            TaskAllocationInfo(
                paasta_service=paasta_service,
                paasta_instance=paasta_instance,
                container_type=MAIN_CONTAINER_TYPE,
                paasta_pool=paasta_pool,
                resources=task["resources"],
                start_time=start_time,
                docker_id=None,
                pod_name=None,
                pod_ip=None,
                host_ip=None,
                git_sha=None,
                config_sha=None,
                mesos_container_id=mesos_container_id,
                namespace=None,
            )
        )
    return info_list


def get_all_running_kubernetes_pods(
    kube_client: KubeClient, namespace: str
) -> Iterable[V1Pod]:
    running = []
    for pod in kubernetes_tools.get_all_pods(kube_client, namespace):
        if kubernetes_tools.get_pod_status(pod) == kubernetes_tools.PodStatus.RUNNING:
            running.append(pod)
    return running


def get_kubernetes_resource_request_limit(
    resources: V1ResourceRequirements,
) -> Dict[str, float]:
    if not resources:
        requests: Dict[str, str] = {}
        limits: Dict[str, str] = {}
    else:
        requests = resources.requests or {}
        limits = resources.limits or {}

    parsed_requests = kubernetes_tools.parse_container_resources(requests)
    parsed_limits = kubernetes_tools.parse_container_resources(limits)
    return {
        "cpus": parsed_requests.cpus,
        "mem": parsed_requests.mem,
        "disk": parsed_requests.disk,
        "cpus_limit": parsed_limits.cpus,
    }


def get_pod_pool(
    kube_client: KubeClient,
    pod: V1Pod,
) -> str:
    node = kubernetes_tools.get_pod_node(kube_client, pod, cache_nodes=True)
    pool = "default"
    if node:
        if node.metadata.labels:
            pool = node.metadata.labels.get("paasta.yelp.com/pool", "default")
    return pool


def get_kubernetes_metadata(
    pod: V1Pod,
) -> Tuple[
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
]:
    labels = pod.metadata.labels or {}
    pod_name = pod.metadata.name
    pod_ip = pod.status.pod_ip
    host_ip = pod.status.host_ip
    service = labels.get("paasta.yelp.com/service")
    instance = labels.get("paasta.yelp.com/instance")
    git_sha = labels.get("paasta.yelp.com/git_sha")
    config_sha = labels.get("paasta.yelp.com/config_sha")
    return service, instance, pod_name, pod_ip, host_ip, git_sha, config_sha


def get_container_type(container_name: str, instance_name: str) -> str:
    """
    To differentiate between main service containers and sidecars
    """
    if instance_name and container_name == kubernetes_tools.sanitise_kubernetes_name(
        instance_name
    ):
        return MAIN_CONTAINER_TYPE
    else:
        return container_name


def get_kubernetes_task_allocation_info(
    namespace: str, client: KubeClient
) -> Iterable[TaskAllocationInfo]:
    pods = get_all_running_kubernetes_pods(client, namespace)
    info_list = []
    for pod in pods:
        (
            service,
            instance,
            pod_name,
            pod_ip,
            host_ip,
            git_sha,
            config_sha,
        ) = get_kubernetes_metadata(pod)
        pool = get_pod_pool(client, pod)
        name_to_info: MutableMapping[str, Any] = {}
        for container in pod.spec.containers:
            name_to_info[container.name] = {
                "resources": get_kubernetes_resource_request_limit(container.resources),
                "container_type": get_container_type(container.name, instance),
                "pod_name": pod_name,
                "pod_ip": pod_ip,
                "host_ip": host_ip,
                "git_sha": git_sha,
                "config_sha": config_sha,
            }
        container_statuses = pod.status.container_statuses or []
        for container in container_statuses:
            if not container.state.running:
                continue
            # docker://abcdef
            docker_id = (
                container.container_id.split("/")[-1]
                if container.container_id
                else None
            )
            update = {
                "docker_id": docker_id,
                "start_time": container.state.running.started_at.timestamp(),
            }
            name_to_info[container.name].update(update)
        for info in name_to_info.values():
            info_list.append(
                TaskAllocationInfo(
                    paasta_service=service,
                    paasta_instance=instance,
                    container_type=info.get("container_type"),
                    paasta_pool=pool,
                    resources=info.get("resources"),
                    start_time=info.get("start_time"),
                    docker_id=info.get("docker_id"),
                    pod_name=info.get("pod_name"),
                    pod_ip=info.get("pod_ip"),
                    host_ip=info.get("host_ip"),
                    git_sha=info.get("git_sha"),
                    config_sha=info.get("config_sha"),
                    mesos_container_id=None,
                    namespace=namespace,
                )
            )

    return info_list


def get_task_allocation_info(
    scheduler: str,
    namespace: str,
    kube_client: Optional[KubeClient],
) -> Iterable[TaskAllocationInfo]:
    if scheduler == "mesos":
        return get_mesos_task_allocation_info()
    elif scheduler == "kubernetes":
        return get_kubernetes_task_allocation_info(namespace, kube_client)
    else:
        return []


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "--scheduler",
        help="Scheduler to get task info from",
        dest="scheduler",
        default="kubernetes",
        choices=["mesos", "kubernetes"],
    )
    parser.add_argument(
        "--additional-namespaces-exclude",
        help="full names of namespaces to not fetch allocation info for those that don't match --namespace-prefix-exlude",
        dest="additional_namespaces_exclude",
        nargs="+",
        default=[],
    )
    parser.add_argument(
        "--namespace-prefix",
        help=argparse.SUPPRESS,
        dest="namespace_prefix",
        default="paasta",
    )
    parser.add_argument(
        "--additional-namespaces",
        help=argparse.SUPPRESS,
        dest="additional_namespaces",
        nargs="+",
        # we default this to tron since this is really the only non-paasta-prefix namespaced that is part of paasta
        # and we'd like to not run two cronjobs to get this information :p
        default=["tron"],
    )
    args = parser.parse_args()

    args.additional_namespaces_exclude = set(args.additional_namespaces_exclude)

    return args


def get_unexcluded_namespaces(
    namespaces: List[str], excluded_namespaces: List[str]
) -> List[str]:
    return [n for n in namespaces if n not in excluded_namespaces]


def main(args: argparse.Namespace) -> None:
    cluster = load_system_paasta_config().get_cluster()
    kube_client = KubeClient()
    all_namespaces = kubernetes_tools.get_all_namespaces(kube_client)
    for matching_namespace in get_unexcluded_namespaces(
        all_namespaces,
        args.additional_namespaces_exclude,
    ):
        display_task_allocation_info(
            cluster, args.scheduler, matching_namespace, kube_client
        )


def display_task_allocation_info(
    cluster: str,
    scheduler: str,
    namespace: str,
    kube_client: Optional[KubeClient],
) -> None:
    info_list = get_task_allocation_info(scheduler, namespace, kube_client)
    timestamp = time.time()
    for info in info_list:
        info_dict = info._asdict()
        info_dict["cluster"] = cluster
        info_dict["timestamp"] = timestamp
        print(json.dumps(info_dict))


if __name__ == "__main__":
    args = parse_args()
    main(args)



from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from unittest import mock

import pytest
from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet
from pytest import raises

from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.kubernetes.application.controller_wrappers import Application
from paasta_tools.kubernetes_tools import HpaOverride
from paasta_tools.kubernetes_tools import InvalidKubernetesConfig
from paasta_tools.kubernetes_tools import KubeDeployment
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfigDict
from paasta_tools.setup_kubernetes_job import create_application_object
from paasta_tools.setup_kubernetes_job import get_kubernetes_deployment_config
from paasta_tools.setup_kubernetes_job import get_service_instances_with_valid_names
from paasta_tools.setup_kubernetes_job import main
from paasta_tools.setup_kubernetes_job import parse_args
from paasta_tools.setup_kubernetes_job import setup_kube_deployments
from paasta_tools.utils import DeploymentVersion
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable


def test_parse_args():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.argparse", autospec=True
    ) as mock_argparse:
        assert parse_args() == mock_argparse.ArgumentParser.return_value.parse_args()


def test_main_logging():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.metrics_lib.get_metrics_interface",
        autospec=True,
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.parse_args", autospec=True
    ) as mock_parse_args, mock.patch(
        "paasta_tools.setup_kubernetes_job.KubeClient", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.get_kubernetes_deployment_config",
        autospec=True,
    ) as mock_service_instance_configs_list, mock.patch(
        "paasta_tools.setup_kubernetes_job.ensure_namespace", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.setup_kube_deployments", autospec=True
    ) as mock_setup_kube_deployments, mock.patch(
        "paasta_tools.setup_kubernetes_job.logging", autospec=True
    ) as mock_logging:
        mock_setup_kube_deployments.return_value = True
        mock_parse_args.return_value.verbose = True
        mock_kube_deploy_config = KubernetesDeploymentConfig(
            service="my-service",
            instance="my-instance",
            cluster="cluster",
            config_dict=KubernetesDeploymentConfigDict(),
            branch_dict=None,
        )
        mock_service_instance_configs_list.return_value = [
            (True, mock_kube_deploy_config)
        ]
        with raises(SystemExit) as e:
            main()
        assert e.value.code == 0
        mock_logging.basicConfig.assert_called_with(level=mock_logging.DEBUG)
        assert not mock_logging.getLogger.called

        mock_parse_args.return_value.verbose = False
        with raises(SystemExit) as e:
            main()
        assert e.value.code == 0
        mock_logging.basicConfig.assert_called_with(level=mock_logging.INFO)
        assert mock_logging.getLogger.called
        mock_logging.getLogger.assert_called_with("kazoo")


@pytest.mark.parametrize(
    "mock_kube_deploy_config, eks_flag",
    [
        (
            KubernetesDeploymentConfig(
                service="my-service",
                instance="my-instance",
                cluster="cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            False,
        ),
        (
            EksDeploymentConfig(
                service="my-service",
                instance="my-instance",
                cluster="cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            True,
        ),
    ],
)
def test_main(mock_kube_deploy_config, eks_flag):
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.metrics_lib.get_metrics_interface",
        autospec=True,
    ) as mock_get_metrics_interface, mock.patch(
        "paasta_tools.setup_kubernetes_job.parse_args", autospec=True
    ) as mock_parse_args, mock.patch(
        "paasta_tools.setup_kubernetes_job.KubeClient", autospec=True
    ) as mock_kube_client, mock.patch(
        "paasta_tools.setup_kubernetes_job.get_kubernetes_deployment_config",
        autospec=True,
    ) as mock_service_instance_configs_list, mock.patch(
        "paasta_tools.setup_kubernetes_job.ensure_namespace", autospec=True
    ) as mock_ensure_namespace, mock.patch(
        "paasta_tools.setup_kubernetes_job.setup_kube_deployments", autospec=True
    ) as mock_setup_kube_deployments, mock.patch(
        "paasta_tools.setup_kubernetes_job.get_hpa_overrides",
        autospec=True,
        return_value={},
    ) as mock_get_hpa_overrides:
        mock_setup_kube_deployments.return_value = True
        mock_metrics_interface = mock_get_metrics_interface.return_value
        mock_parse_args.return_value.eks = eks_flag
        mock_service_instance_configs_list.return_value = [
            (True, mock_kube_deploy_config)
        ]
        with raises(SystemExit) as e:
            main()
        assert e.value.code == 0
        assert mock_ensure_namespace.called
        mock_setup_kube_deployments.assert_called_with(
            kube_client=mock_kube_client.return_value,
            cluster=mock_parse_args.return_value.cluster,
            soa_dir=mock_parse_args.return_value.soa_dir,
            rate_limit=mock_parse_args.return_value.rate_limit,
            service_instance_configs_list=mock_service_instance_configs_list.return_value,
            metrics_interface=mock_metrics_interface,
            eks=mock_parse_args.return_value.eks,
            hpa_overrides=mock_get_hpa_overrides.return_value,
        )
        mock_setup_kube_deployments.return_value = False
        with raises(SystemExit) as e:
            main()
        assert e.value.code == 1

        mock_service_instance_configs_list.return_value = [(False, None)]
        with raises(SystemExit) as e:
            main()
        assert e.value.code == 1


def test_get_service_instances_with_valid_names():
    mock_service_instances = ["kurupt.f_m"]
    ret = get_service_instances_with_valid_names(
        service_instances=mock_service_instances
    )

    assert ret == [("kurupt", "f_m", None, None)]


def test_main_invalid_job_name():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.parse_args", autospec=True
    ) as mock_parse_args, mock.patch(
        "paasta_tools.setup_kubernetes_job.KubeClient", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.metrics_lib.get_metrics_interface",
        autospec=True,
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.load_system_paasta_config", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.create_application_object", autospec=True
    ) as mock_create_application_object:
        mock_parse_args.return_value.cluster = "fake_cluster"
        mock_parse_args.return_value.soa_dir = "/etc/fake"
        mock_parse_args.return_value.service_instance_list = ["kuruptf_m"]
        with raises(SystemExit) as e:
            main()
        assert mock_create_application_object.call_count == 0
        assert e.value.code == 1


def test_get_kubernetes_deployment_config():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.load_kubernetes_service_config_no_cache",
        autospec=True,
    ) as mock_load_kubernetes_service_config_no_cache:

        mock_get_service_instances_with_valid_names = [
            ("kurupt", "instance", None, None)
        ]

        # Testing NoDeploymentsAvailable exception
        mock_load_kubernetes_service_config_no_cache.side_effect = (
            NoDeploymentsAvailable
        )
        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
        )
        assert ret == [(True, None)]

        # Testing NoConfigurationForServiceError exception
        mock_load_kubernetes_service_config_no_cache.side_effect = (
            NoConfigurationForServiceError
        )

        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
        )
        assert ret == [(False, None)]

        # Testing returning a KubernetesDeploymentConfig
        mock_kube_deploy = KubernetesDeploymentConfig(
            service="kurupt",
            instance="instance",
            cluster="fake_cluster",
            soa_dir="nail/blah",
            config_dict=KubernetesDeploymentConfigDict(),
            branch_dict=None,
        )
        mock_load_kubernetes_service_config_no_cache.side_effect = None
        mock_load_kubernetes_service_config_no_cache.return_value = mock_kube_deploy
        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
        )

        assert ret == [
            (
                True,
                KubernetesDeploymentConfig(
                    service="kurupt",
                    instance="instance",
                    cluster="fake_cluster",
                    soa_dir="nail/blah",
                    config_dict=KubernetesDeploymentConfigDict(),
                    branch_dict=None,
                ),
            )
        ]


def test_get_eks_deployment_config():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.load_eks_service_config_no_cache",
        autospec=True,
    ) as mock_load_eks_service_config_no_cache:

        mock_get_service_instances_with_valid_names = [
            ("kurupt", "instance", None, None)
        ]

        # Testing NoDeploymentsAvailable exception
        mock_load_eks_service_config_no_cache.side_effect = NoDeploymentsAvailable
        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
            eks=True,
        )
        assert ret == [(True, None)]

        # Testing NoConfigurationForServiceError exception
        mock_load_eks_service_config_no_cache.side_effect = (
            NoConfigurationForServiceError
        )

        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
            eks=True,
        )
        assert ret == [(False, None)]

        # Testing returning a KubernetesDeploymentConfig
        mock_kube_deploy = EksDeploymentConfig(
            service="kurupt",
            instance="instance",
            cluster="fake_cluster",
            soa_dir="nail/blah",
            config_dict=KubernetesDeploymentConfigDict(),
            branch_dict=None,
        )
        mock_load_eks_service_config_no_cache.side_effect = None
        mock_load_eks_service_config_no_cache.return_value = mock_kube_deploy
        ret = get_kubernetes_deployment_config(
            service_instances_with_valid_names=mock_get_service_instances_with_valid_names,
            cluster="fake_cluster",
            soa_dir="nail/blah",
            eks=True,
        )

        assert ret == [
            (
                True,
                EksDeploymentConfig(
                    service="kurupt",
                    instance="instance",
                    cluster="fake_cluster",
                    soa_dir="nail/blah",
                    config_dict=KubernetesDeploymentConfigDict(),
                    branch_dict=None,
                ),
            )
        ]


@pytest.mark.parametrize(
    "eks_flag, mock_service_config",
    [
        (
            "False",
            KubernetesDeploymentConfig(
                service="kurupt",
                instance="instance",
                cluster="fake_cluster",
                soa_dir="nail/blah",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
        ),
        (
            "True",
            EksDeploymentConfig(
                service="kurupt",
                instance="instance",
                cluster="fake_cluster",
                soa_dir="nail/blah",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
        ),
    ],
)
def test_create_application_object(eks_flag, mock_service_config):
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.load_system_paasta_config", autospec=True
    ), mock.patch(
        "paasta_tools.kubernetes.application.controller_wrappers.Application.load_local_config",
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes.application.controller_wrappers.DeploymentWrapper",
        autospec=True,
    ) as mock_deployment_wrapper, mock.patch(
        "paasta_tools.kubernetes.application.controller_wrappers.StatefulSetWrapper",
        autospec=True,
    ) as mock_stateful_set_wrapper:
        mock_deploy = mock.MagicMock(spec=V1Deployment)
        service_config = mock.MagicMock(spec=mock_service_config)
        service_config.format_kubernetes_app.return_value = mock_deploy

        # Create DeploymentWrapper
        create_application_object(
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            service_instance_config=service_config,
            eks=eks_flag,
        )

        mock_deployment_wrapper.assert_called_with(mock_deploy, hpa_override=None)

        mock_deploy = mock.MagicMock(spec=V1StatefulSet)
        service_config.format_kubernetes_app.return_value = mock_deploy

        # Create StatefulSetWrapper
        create_application_object(
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            service_instance_config=service_config,
            eks=eks_flag,
        )
        mock_stateful_set_wrapper.assert_called_with(mock_deploy)

        # Create object that is not statefulset/deployment
        with raises(Exception):
            service_config.format_kubernetes_app.return_value = mock.MagicMock()
            create_application_object(
                cluster="fake_cluster",
                soa_dir="/nail/blah",
                service_instance_config=service_config,
                eks=eks_flag,
            )

        mock_deployment_wrapper.reset_mock()
        mock_stateful_set_wrapper.reset_mock()

        service_config.format_kubernetes_app.side_effect = InvalidKubernetesConfig(
            Exception("Oh no!"), "kurupt", "fm"
        )
        ret = create_application_object(
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            service_instance_config=service_config,
            eks=eks_flag,
        )

        assert ret == (False, None)
        assert not mock_deployment_wrapper.called
        assert not mock_stateful_set_wrapper.called


@pytest.mark.parametrize(
    "mock_kube_deploy_config, eks_flag",
    [
        (
            KubernetesDeploymentConfig(
                service="kurupt",
                instance="fm",
                cluster="fake_cluster",
                soa_dir="/nail/blah",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            False,
        ),
        (
            EksDeploymentConfig(
                service="kurupt",
                instance="fm",
                cluster="fake_cluster",
                soa_dir="/nail/blah",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            True,
        ),
    ],
)
def test_setup_kube_deployment_create_update(mock_kube_deploy_config, eks_flag):
    fake_create = mock.MagicMock()
    fake_update = mock.MagicMock()
    fake_update_related_api_objects = mock.MagicMock()

    def simple_create_application_object(
        cluster,
        soa_dir,
        service_instance_config,
        eks,
        hpa_override,
    ):
        fake_app = mock.MagicMock(spec=Application)
        fake_app.kube_deployment = KubeDeployment(
            service=service_instance_config.service,
            instance=service_instance_config.instance,
            namespace="paasta",
            git_sha="1",
            image_version="extrastuff-1",
            config_sha="1",
            replicas=1,
        )
        fake_app.create = fake_create
        fake_app.update = fake_update
        fake_app.update_related_api_objects = fake_update_related_api_objects
        fake_app.item = None
        fake_app.soa_config = KubernetesDeploymentConfig(
            service=service_instance_config.service,
            cluster=cluster,
            instance=service_instance_config.instance,
            config_dict=service_instance_config.config_dict,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        fake_app.__str__ = lambda app: "fake_app"  # type: ignore
        return True, fake_app

    with mock.patch(
        "paasta_tools.setup_kubernetes_job.create_application_object",
        autospec=True,
        side_effect=simple_create_application_object,
    ) as mock_create_application_object, mock.patch(
        "paasta_tools.setup_kubernetes_job.list_all_paasta_deployments", autospec=True
    ) as mock_list_all_paasta_deployments, mock.patch(
        "paasta_tools.setup_kubernetes_job.log", autospec=True
    ) as mock_log_obj, mock.patch(
        "paasta_tools.setup_kubernetes_job.metrics_lib.NoMetrics", autospec=True
    ) as mock_no_metrics, mock.patch(
        "paasta_tools.setup_kubernetes_job.get_kubernetes_deployment_config",
        autospec=True,
    ):
        mock_client = mock.Mock()
        # No instances created
        mock_service_instance_configs_list: List[
            Tuple[bool, Union[KubernetesDeploymentConfig, EksDeploymentConfig]]
        ] = []
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
        )
        assert mock_create_application_object.call_count == 0
        assert fake_update.call_count == 0
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 0
        assert mock_log_obj.info.call_count == 0
        mock_log_obj.info.reset_mock()

        # Create a new instance
        mock_service_instance_configs_list = [(True, mock_kube_deploy_config)]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            metrics_interface=mock_no_metrics,
            eks=eks_flag,
        )
        assert fake_create.call_count == 1
        assert fake_update.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        assert mock_no_metrics.emit_event.call_count == 1
        mock_log_obj.info.reset_mock()
        mock_no_metrics.reset_mock()

        # Skipping downthenup instance cuz of existing_apps
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="2",
                namespace="paastasvc-kurupt",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=1,
            )
        ]
        mock_downthenup_kube_deploy_config = KubernetesDeploymentConfig(
            service="kurupt",
            instance="fm",
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            config_dict=KubernetesDeploymentConfigDict(bounce_method="downthenup"),
            branch_dict=None,
        )
        mock_service_instance_configs_list = [
            (True, mock_downthenup_kube_deploy_config)
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            metrics_interface=mock_no_metrics,
            eks=eks_flag,
        )
        assert fake_create.call_count == 0
        assert fake_update.call_count == 0
        assert fake_update_related_api_objects.call_count == 0
        assert mock_no_metrics.emit_event.call_count == 0
        mock_log_obj.info.reset_mock()
        mock_no_metrics.reset_mock()

        # Update when gitsha changed
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="2",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=1,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            metrics_interface=mock_no_metrics,
            eks=eks_flag,
        )

        assert fake_update.call_count == 1
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        mock_no_metrics.emit_event.assert_called_with(
            name="deploy",
            dimensions={
                "paasta_cluster": "fake_cluster",
                "paasta_service": "kurupt",
                "paasta_instance": "fm",
                "paasta_namespace": "paasta",
                "deploy_event": "update",
            },
        )
        mock_log_obj.info.reset_mock()
        mock_no_metrics.reset_mock()

        # Update when image_version changed
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-2",
                config_sha="1",
                replicas=1,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            eks=eks_flag,
        )
        assert fake_update.call_count == 1
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        mock_log_obj.info.reset_mock()

        # Update when configsha changed
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="2",
                replicas=1,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            eks=eks_flag,
        )
        assert fake_update.call_count == 1
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        mock_log_obj.info.reset_mock()

        # Update when replica changed
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=2,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            eks=eks_flag,
        )
        assert fake_update.call_count == 1
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        mock_log_obj.info.reset_mock()

        # Update one and Create One
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_kube_deploy_config_new = KubernetesDeploymentConfig(
            service="kurupt",
            instance="garage",
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            config_dict=KubernetesDeploymentConfigDict(),
            branch_dict=None,
        )
        mock_service_instance_configs_list = [
            (True, mock_kube_deploy_config),
            (True, mock_kube_deploy_config_new),
        ]
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="garage",
                git_sha="2",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="2",
                replicas=1,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            eks=eks_flag,
        )
        assert fake_update.call_count == 1
        assert fake_create.call_count == 1
        assert fake_update_related_api_objects.call_count == 2
        mock_log_obj.info.reset_mock()

        # Always attempt to update related API objects
        fake_create.reset_mock()
        fake_update.reset_mock()
        fake_update_related_api_objects.reset_mock()
        mock_service_instance_configs_list = [(True, mock_kube_deploy_config_new)]
        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="garage",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=1,
            )
        ]
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            eks=eks_flag,
        )
        assert fake_update.call_count == 0
        assert fake_create.call_count == 0
        assert fake_update_related_api_objects.call_count == 1
        assert mock_log_obj.info.call_args_list[0] == mock.call(
            "fake_app is up-to-date!"
        )


@pytest.mark.parametrize(
    "mock_kube_deploy_config_fm, mock_kube_deploy_config_garage, mock_kube_deploy_config_radio, eks_flag",
    [
        (
            KubernetesDeploymentConfig(
                service="kurupt",
                instance="fm",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            KubernetesDeploymentConfig(
                service="kurupt",
                instance="garage",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            KubernetesDeploymentConfig(
                service="kurupt",
                instance="radio",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            False,
        ),
        (
            EksDeploymentConfig(
                service="kurupt",
                instance="fm",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            EksDeploymentConfig(
                service="kurupt",
                instance="garage",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            EksDeploymentConfig(
                service="kurupt",
                instance="radio",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            True,
        ),
    ],
)
def test_setup_kube_deployments_rate_limit(
    mock_kube_deploy_config_fm,
    mock_kube_deploy_config_garage,
    mock_kube_deploy_config_radio,
    eks_flag,
):
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.create_application_object",
        autospec=True,
    ) as mock_create_application_object, mock.patch(
        "paasta_tools.setup_kubernetes_job.list_all_paasta_deployments", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.log", autospec=True
    ) as mock_log_obj:
        mock_client = mock.Mock()
        mock_service_instance_configs_list: List[
            Tuple[bool, Union[KubernetesDeploymentConfig, EksDeploymentConfig]]
        ] = [
            (True, mock_kube_deploy_config_fm),
            (True, mock_kube_deploy_config_garage),
            (True, mock_kube_deploy_config_radio),
        ]
        fake_app = mock.Mock(create=mock.Mock())
        mock_create_application_object.return_value = (True, fake_app)

        # Rate limit: 2 calls allowed
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            rate_limit=2,
            eks=eks_flag,
        )
        assert fake_app.create.call_count == 2
        mock_log_obj.info.assert_any_call(
            "Not doing any further updates as we reached the limit (2)"
        )

        # No rate limit
        fake_app.reset_mock()
        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            rate_limit=0,
            eks=eks_flag,
        )
        assert fake_app.create.call_count == 3


@pytest.mark.parametrize(
    "mock_kube_deploy_config_fake, mock_kube_deploy_config_mock, eks_flag",
    [
        (
            KubernetesDeploymentConfig(
                service="fake",
                instance="instance",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            KubernetesDeploymentConfig(
                service="mock",
                instance="instance",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            False,
        ),
        (
            EksDeploymentConfig(
                service="fake",
                instance="instance",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            EksDeploymentConfig(
                service="mock",
                instance="instance",
                cluster="fake_cluster",
                config_dict=KubernetesDeploymentConfigDict(),
                branch_dict=None,
            ),
            True,
        ),
    ],
)
def test_setup_kube_deployments_skip_malformed_apps(
    mock_kube_deploy_config_fake, mock_kube_deploy_config_mock, eks_flag
):
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.create_application_object",
        autospec=True,
    ) as mock_create_application_object, mock.patch(
        "paasta_tools.setup_kubernetes_job.list_all_paasta_deployments", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_job.log", autospec=True
    ) as mock_log_obj:
        mock_client = mock.Mock()
        mock_service_instance_configs_list: List[
            Tuple[bool, Union[KubernetesDeploymentConfig, EksDeploymentConfig]]
        ] = [
            (True, mock_kube_deploy_config_fake),
            (True, mock_kube_deploy_config_mock),
        ]
        fake_app = mock.Mock(create=mock.Mock())
        fake_app.create = mock.Mock(
            side_effect=[Exception("Kaboom!"), mock.Mock(create=mock.Mock())]
        )
        fake_app.__str__ = mock.Mock(return_value="fake_app")  # type: ignore
        mock_create_application_object.return_value = (True, fake_app)

        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            rate_limit=0,
            eks=eks_flag,
        )
        assert fake_app.create.call_count == 2
        assert len(mock_log_obj.exception.call_args_list) == 1
        assert mock_log_obj.exception.call_args_list[0] == mock.call(
            "Error while processing: fake_app"
        )


def test_setup_kube_deployments_does_git_sha_changes_first():
    with mock.patch(
        "paasta_tools.setup_kubernetes_job.create_application_object",
        autospec=True,
    ) as mock_create_application_object, mock.patch(
        "paasta_tools.setup_kubernetes_job.list_all_paasta_deployments", autospec=True
    ) as mock_list_all_paasta_deployments, mock.patch(
        "paasta_tools.setup_kubernetes_job.log", autospec=True
    ) as mock_log_obj:
        mock_client = mock.Mock()
        mock_kube_deploy_config_fm = KubernetesDeploymentConfig(
            service="kurupt",
            instance="fm",
            cluster="fake_cluster",
            config_dict={},
            branch_dict=None,
        )
        mock_kube_deploy_config_garage = KubernetesDeploymentConfig(
            service="kurupt",
            instance="garage",
            cluster="fake_cluster",
            config_dict={},
            branch_dict=None,
        )

        mock_list_all_paasta_deployments.return_value = [
            KubeDeployment(
                service="kurupt",
                instance="garage",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=1,
            ),
            KubeDeployment(
                service="kurupt",
                instance="fm",
                git_sha="1",
                namespace="paasta",
                image_version="extrastuff-1",
                config_sha="1",
                replicas=1,
            ),
        ]

        mock_service_instance_configs_list = [
            (True, mock_kube_deploy_config_garage),
            (True, mock_kube_deploy_config_fm),
        ]

        garage_app = mock.Mock(
            kube_deployment=mock.Mock(
                git_sha="1",
                service="kurupt",
                instance="garage",
                image_version="extrastuff-1",
                namespace="paasta",
                deployment_version=DeploymentVersion("1", "extrastuff-1"),
            ),
        )
        fm_app = mock.Mock(
            kube_deployment=mock.Mock(
                git_sha="2",
                service="kurupt",
                instance="fm",
                image_version="extrastuff-1",
                namespace="paasta",
                deployment_version=DeploymentVersion("2", "extrastuff-1"),
            ),
        )

        def fake_create_application_object(
            cluster: str,
            soa_dir: str,
            service_instance_config: KubernetesDeploymentConfig,
            eks: bool = False,
            hpa_override: Optional[HpaOverride] = None,
        ):
            if service_instance_config.instance == "garage":
                return (True, garage_app)
            elif service_instance_config.instance == "fm":
                return (True, fm_app)
            raise ValueError("expecting instance of 'fm' or 'garage'")

        mock_create_application_object.side_effect = fake_create_application_object

        setup_kube_deployments(
            kube_client=mock_client,
            service_instance_configs_list=mock_service_instance_configs_list,
            cluster="fake_cluster",
            soa_dir="/nail/blah",
            rate_limit=1,
        )

        # With rate_limit == 1, only the app with a different `git_sha` should be bounced.
        assert garage_app.update.call_count == 0
        assert fm_app.update.call_count == 1
        mock_log_obj.info.assert_any_call(
            "Not doing any further updates as we reached the limit (1)"
        )



# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
from unittest import mock

import pytest
import requests

from paasta_tools.envoy_tools import are_namespaces_up_in_eds
from paasta_tools.envoy_tools import are_services_up_in_pod
from paasta_tools.envoy_tools import get_backends
from paasta_tools.envoy_tools import get_backends_from_eds
from paasta_tools.envoy_tools import get_casper_endpoints
from paasta_tools.envoy_tools import match_backends_and_pods


def test_get_backends():
    testdir = os.path.dirname(os.path.realpath(__file__))
    testdata = os.path.join(testdir, "envoy_admin_clusters_snapshot.txt")
    with open(testdata, "r") as fd:
        mock_envoy_admin_clusters_data = json.load(fd)

    mock_response = mock.Mock()
    mock_response.json.return_value = mock_envoy_admin_clusters_data
    mock_get = mock.Mock(return_value=(mock_response))

    hosts = {
        "10.46.6.90": ("host2.two.com", None, None),
        "10.46.6.88": ("host3.three.com", None, None),
        "10.46.6.103": ("host4.four.com", None, None),
    }

    with mock.patch.object(requests.Session, "get", mock_get):
        with mock.patch(
            "socket.gethostbyaddr",
            side_effect=lambda x: hosts[x],
            autospec=True,
        ):
            expected = {
                "service1.main": [
                    (
                        {
                            "address": "10.46.6.88",
                            "port_value": 13833,
                            "hostname": "host3",
                            "eds_health_status": "HEALTHY",
                            "weight": 1,
                        },
                        False,
                    ),
                    (
                        {
                            "address": "10.46.6.90",
                            "port_value": 13833,
                            "hostname": "host2",
                            "eds_health_status": "HEALTHY",
                            "weight": 1,
                        },
                        False,
                    ),
                ]
            }
            assert expected == get_backends("service1.main", "host", 123, "something")


def test_get_casper_endpoints():
    testdir = os.path.dirname(os.path.realpath(__file__))
    testdata = os.path.join(testdir, "envoy_admin_clusters_snapshot.txt")
    with open(testdata, "r") as fd:
        mock_envoy_admin_clusters_data = json.load(fd)

    expected = frozenset([("10.46.6.106", 13819)])

    assert expected == get_casper_endpoints(mock_envoy_admin_clusters_data)


@pytest.fixture
def mock_backends():
    return [
        {
            "address": "10.50.2.4",
            "port_value": 31000,
            "eds_health_status": "HEALTHY",
            "weight": 1,
            "has_associated_task": False,
        },
        {
            "address": "10.50.2.5",
            "port_value": 31001,
            "eds_health_status": "HEALTHY",
            "weight": 1,
            "has_associated_task": False,
        },
        {
            "address": "10.50.2.6",
            "port_value": 31001,
            "eds_health_status": "HEALTHY",
            "weight": 1,
            "has_associated_task": False,
        },
        {
            "address": "10.50.2.6",
            "port_value": 31002,
            "eds_health_status": "HEALTHY",
            "weight": 1,
            "has_associated_task": False,
        },
        {
            "address": "10.50.2.8",
            "port_value": 31000,
            "eds_health_status": "HEALTHY",
            "weight": 1,
            "has_associated_task": False,
        },
    ]


def test_match_backends_and_pods(mock_backends):
    backends = mock_backends

    good_pod_1 = mock.Mock(status=mock.Mock(pod_ip="10.50.2.4"))
    good_pod_2 = mock.Mock(status=mock.Mock(pod_ip="10.50.2.5"))
    bad_pod_1 = mock.Mock(status=mock.Mock(pod_ip="10.50.2.10"))
    pods = [good_pod_1, good_pod_2, bad_pod_1]

    expected = [
        (backends[0], good_pod_1),
        (backends[1], good_pod_2),
        (None, bad_pod_1),
        (backends[2], None),
        (backends[3], None),
        (backends[4], None),
    ]
    actual = match_backends_and_pods(backends, pods)

    def keyfunc(t):
        sorted_backend = tuple(sorted((t[0] or {}).items()))
        pod_ip = t[1].status.pod_ip if t[1] else ""
        return sorted_backend, pod_ip

    assert sorted(actual, key=keyfunc) == sorted(expected, key=keyfunc)


class TestServicesUpInPod:
    pod_ip = "10.40.1.1"
    pod_port = 8888

    @pytest.fixture
    def cluster(self):
        def _make_cluster(health, ip=self.pod_ip):
            return (
                {
                    "eds_health_status": health,
                    "address": ip,
                    "port_value": self.pod_port,
                },
                False,
            )

        return _make_cluster

    @pytest.fixture
    def mock_get_multiple_backends(self):
        with mock.patch(
            "paasta_tools.envoy_tools.get_multiple_backends", autospec=True
        ) as mock_get_multiple_backends:
            yield mock_get_multiple_backends

    def test_are_services_up_on_port_no_clusters(self, mock_get_multiple_backends):
        mock_get_multiple_backends.return_value = {}
        assert not are_services_up_in_pod(
            envoy_host="1.2.3.4",
            envoy_admin_port=3212,
            envoy_admin_endpoint_format="http://{bla}:{more_bla}",
            registrations=["service1.instance1", "service1.instance2"],
            pod_ip=self.pod_ip,
            pod_port=self.pod_port,
        )

    def test_are_services_up_on_port_all_backends_healthy(
        self, mock_get_multiple_backends, cluster
    ):
        mock_get_multiple_backends.side_effect = [
            {
                "service1.instance1": [
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
            {
                "service1.instance2": [
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
        ]
        assert are_services_up_in_pod(
            envoy_host="1.2.3.4",
            envoy_admin_port=3212,
            envoy_admin_endpoint_format="http://{bla}:{more_bla}",
            registrations=["service1.instance1", "service1.instance2"],
            pod_ip=self.pod_ip,
            pod_port=self.pod_port,
        )

    def test_are_services_up_on_port_unhealthy_service(
        self, mock_get_multiple_backends, cluster
    ):
        mock_get_multiple_backends.side_effect = [
            {
                "service1.instance1": [
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
            {
                "service1.instance2": [
                    cluster("UNHEALTHY"),
                    cluster("UNHEALTHY"),
                    cluster("UNHEALTHY"),
                ]
            },
        ]
        assert not are_services_up_in_pod(
            envoy_host="1.2.3.4",
            envoy_admin_port=3212,
            envoy_admin_endpoint_format="http://{bla}:{more_bla}",
            registrations=["service1.instance1", "service1.instance2"],
            pod_ip=self.pod_ip,
            pod_port=self.pod_port,
        )

    def test_are_services_up_on_port_partial_health_backend(
        self, mock_get_multiple_backends, cluster
    ):
        mock_get_multiple_backends.return_value = [
            cluster("HEALTHY"),
            cluster("HEALTHY"),
            cluster("UNHEALTHY"),
        ]
        mock_get_multiple_backends.side_effect = [
            {
                "service1.instance1": [
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                    cluster("UNHEALTHY"),
                ]
            },
            {
                "service1.instance2": [
                    cluster("HEALTHY"),
                    cluster("UNHEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
        ]
        assert are_services_up_in_pod(
            envoy_host="1.2.3.4",
            envoy_admin_port=3212,
            envoy_admin_endpoint_format="http://{bla}:{more_bla}",
            registrations=["service1.instance1", "service1.instance2"],
            pod_ip=self.pod_ip,
            pod_port=self.pod_port,
        )

    def test_are_services_up_on_port_missing_backend(
        self, mock_get_multiple_backends, cluster
    ):
        mock_get_multiple_backends.side_effect = [
            [cluster("HEALTHY"), cluster("HEALTHY"), cluster("HEALTHY")],
            [cluster("HEALTHY"), cluster("HEALTHY"), cluster("HEALTHY")],
            [],
        ]
        mock_get_multiple_backends.side_effect = [
            {
                "service1.instance1": [
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
            {
                "service1.instance2": [
                    cluster("HEALTHY"),
                    cluster("UNHEALTHY"),
                    cluster("HEALTHY"),
                ]
            },
            {},
        ]
        # all up and present but service1.instance3 not present
        assert not are_services_up_in_pod(
            envoy_host="1.2.3.4",
            envoy_admin_port=3212,
            envoy_admin_endpoint_format="http://{bla}:{more_bla}",
            registrations=[
                "service1.instance1",
                "service1.instance2",
                "service1.instance3",
            ],
            pod_ip=self.pod_ip,
            pod_port=self.pod_port,
        )

    @mock.patch("paasta_tools.envoy_tools.open", autospec=False)
    @mock.patch("paasta_tools.envoy_tools.os.access", autospec=True)
    @mock.patch("paasta_tools.envoy_tools.yaml.safe_load", autospec=True)
    def test_get_backends_from_eds(self, mock_yaml, mock_os_access, mock_open):

        mock_yaml.return_value = {"resources": [{"endpoints": None}]}
        backends = get_backends_from_eds("my-namespace", "/var/bla")
        assert len(backends) == 0

        mock_yaml.return_value = {
            "resources": [
                {
                    "endpoints": [
                        {
                            "priority": 0,
                            "lb_endpoints": [
                                {
                                    "endpoint": {
                                        "address": {
                                            "socket_address": {
                                                "address": "1.2.3.4",
                                                "port_value": 5555,
                                            }
                                        }
                                    }
                                },
                                {
                                    "endpoint": {
                                        "address": {
                                            "socket_address": {
                                                "address": "5.6.7.8",
                                                "port_value": 5555,
                                            }
                                        }
                                    }
                                },
                            ],
                        },
                        {
                            "priority": 1,
                            "lb_endpoints": [
                                {
                                    "endpoint": {
                                        "address": {
                                            "socket_address": {
                                                "address": "9.10.11.12",
                                                "port_value": 5555,
                                            }
                                        }
                                    }
                                },
                                {
                                    "endpoint": {
                                        "address": {
                                            "socket_address": {
                                                "address": "13.14.15.16",
                                                "port_value": 5555,
                                            }
                                        }
                                    }
                                },
                            ],
                        },
                    ]
                }
            ]
        }
        expected_backends = [
            ("1.2.3.4", 5555),
            ("5.6.7.8", 5555),
            ("9.10.11.12", 5555),
            ("13.14.15.16", 5555),
        ]
        backends = get_backends_from_eds("my-namespace", "/var/bla")
        assert sorted(backends) == sorted(expected_backends)

    @mock.patch("paasta_tools.envoy_tools.get_backends_from_eds", autospec=True)
    def test_are_namespaces_up_in_eds(self, mock_get_backends_from_eds):
        # No backends for service1.instance1 and service1.instance2
        mock_get_backends_from_eds.return_value = []
        assert not are_namespaces_up_in_eds(
            envoy_eds_path="/eds/path",
            namespaces=["service1.instance1", "service1.instance2"],
            pod_ip="1.2.3.4",
            pod_port=50000,
        )

        # No backends for service1.instance2
        mock_get_backends_from_eds.side_effect = [
            [("1.2.3.4", 50000), ("5.6.7.8", 60000)],
            [],
        ]
        assert not are_namespaces_up_in_eds(
            envoy_eds_path="/eds/path",
            namespaces=["service1.instance1", "service1.instance2"],
            pod_ip="1.2.3.4",
            pod_port=50000,
        )

        # Missing backend for service1.instance2
        mock_get_backends_from_eds.side_effect = [
            [("1.2.3.4", 50000), ("5.6.7.8", 60000)],
            [("5.6.7.8", 60000)],
        ]
        assert not are_namespaces_up_in_eds(
            envoy_eds_path="/eds/path",
            namespaces=["service1.instance1", "service1.instance2"],
            pod_ip="1.2.3.4",
            pod_port=50000,
        )

        # Backends returned for both namespaces
        mock_get_backends_from_eds.side_effect = [
            [("1.2.3.4", 50000), ("5.6.7.8", 60000)],
            [("5.6.7.8", 60000), ("1.2.3.4", 50000)],
        ]
        assert are_namespaces_up_in_eds(
            envoy_eds_path="/eds/path",
            namespaces=["service1.instance1", "service1.instance2"],
            pod_ip="1.2.3.4",
            pod_port=50000,
        )



import os
from unittest import mock

import pytest

import paasta_tools.config_utils as config_utils
from paasta_tools import yaml_tools as yaml
from paasta_tools.utils import AUTO_SOACONFIG_SUBDIR


@pytest.fixture(autouse=True)
def mock_subprocess():
    with mock.patch(
        "paasta_tools.config_utils.subprocess", autospec=True
    ) as mock_subprocess:
        yield mock_subprocess


@pytest.fixture
def updater(tmpdir):
    updater = config_utils.AutoConfigUpdater(
        "test_source", "remote", working_dir=tmpdir
    )

    with updater:
        # make a valid service dir we can write in
        os.mkdir(os.path.join(updater.working_dir, "foo"))
        yield updater


def test_write_auto_config_data_service_dne(tmpdir):
    with mock.patch(
        "paasta_tools.config_utils.open",
        new=mock.mock_open(),
        autospec=None,
    ) as mock_open:
        result = config_utils.write_auto_config_data(
            service="something",
            extra_info="kubernetes-norcal-devc",
            data={"a": 1},
            soa_dir=tmpdir,
        )
        assert len(mock_open.mock_calls) == 0
    assert result is None


def test_write_auto_config_data_new_file(tmpdir):
    service = "foo"
    conf_file = "kubernetes-norcal-devc"
    data = {"a": 1}

    tmpdir.mkdir(service)
    result = config_utils.write_auto_config_data(
        service=service,
        extra_info=conf_file,
        data=data,
        soa_dir=tmpdir,
        sub_dir=AUTO_SOACONFIG_SUBDIR,
    )
    expected_path = f"{tmpdir}/{service}/{AUTO_SOACONFIG_SUBDIR}/{conf_file}.yaml"
    assert result == expected_path
    with open(expected_path) as f:
        assert yaml.safe_load(f) == data


def test_write_auto_config_data_file_exists(tmpdir):
    service = "foo"
    conf_file = "kubernetes-norcal-devc"

    tmpdir.mkdir(service)
    config_utils.write_auto_config_data(
        service=service,
        extra_info=conf_file,
        data={"a": 1},
        soa_dir=tmpdir,
        sub_dir=AUTO_SOACONFIG_SUBDIR,
    )
    # Contents should be replaced on second write
    result = config_utils.write_auto_config_data(
        service=service,
        extra_info=conf_file,
        data={"a": 2},
        soa_dir=tmpdir,
        sub_dir=AUTO_SOACONFIG_SUBDIR,
    )
    expected_path = f"{tmpdir}/{service}/{AUTO_SOACONFIG_SUBDIR}/{conf_file}.yaml"
    assert result == expected_path
    with open(expected_path) as f:
        assert yaml.safe_load(f) == {"a": 2}


@mock.patch("paasta_tools.config_utils.validate_schema", autospec=True)
def test_validate_auto_config_file_config_types(mock_validate, tmpdir):
    for config_type in (
        "kubernetes",
        "deploy",
        "smartstack",
        "cassandracluster",
    ):
        filepath = f"service/{config_type}-cluster.yaml"
        assert config_utils.validate_auto_config_file(filepath, AUTO_SOACONFIG_SUBDIR)
        mock_validate.assert_called_with(filepath, f"autotuned_defaults/{config_type}")


@mock.patch("paasta_tools.config_utils.validate_schema", autospec=True)
def test_validate_auto_config_file_unknown_type(mock_validate, tmpdir):
    assert not config_utils.validate_auto_config_file(
        "service/unknown-thing.yaml", AUTO_SOACONFIG_SUBDIR
    )


@pytest.mark.parametrize(
    "data,is_valid",
    [
        ({"not_an_instance": 1}, False),
        ({"instance": {"cpus": "bad_string_value"}}, False),
        ({"instance": {"cpus": 1, "invalid_key": 2}}, False),
        ({"instance": {"cpus": 1.2, "mem": 100}}, True),
        ({"instance": {"cpus": 1.2, "mem": 100, "disk": "very big"}}, False),
        ({"instance": {"cpus": 1.2, "mem": 100, "disk": 0}}, False),
        ({"instance": {"cpus": 1.2, "mem": 100, "disk": 1000}}, True),
        ({"instance": {"min_instances": 1.2}}, False),
        ({"instance": {"min_instances": 1}}, True),
    ],
)
def test_validate_auto_config_file_e2e(data, is_valid, tmpdir):
    service = "foo"
    conf_file = "kubernetes-norcal-devc"

    tmpdir.mkdir(service)
    filepath = config_utils.write_auto_config_data(
        service=service,
        extra_info=conf_file,
        data=data,
        soa_dir=tmpdir,
    )
    assert (
        config_utils.validate_auto_config_file(filepath, AUTO_SOACONFIG_SUBDIR)
        == is_valid
    )


@pytest.mark.parametrize(
    "branch, remote_branch_exists",
    [("master", True), ("other_test", True), ("other_test", False)],
)
def test_auto_config_updater_context(
    branch, remote_branch_exists, tmpdir, mock_subprocess
):
    remote = "git_remote"
    updater = config_utils.AutoConfigUpdater(
        "test_source", remote, branch=branch, working_dir=tmpdir
    )
    updater._remote_branch_exists = mock.MagicMock(
        autospec=True, return_value=remote_branch_exists
    )
    initial_wd = os.getcwd()

    with updater:
        clone_dir = updater.working_dir
        assert os.path.isdir(clone_dir)
        expected_calls = [mock.call.check_call(["git", "clone", remote, clone_dir])]
        if branch != "master":
            if remote_branch_exists:
                expected_calls.extend(
                    [
                        mock.call.check_call(["git", "fetch", "origin", branch]),
                        mock.call.check_call(
                            ["git", "checkout", "-b", branch, f"origin/{branch}"]
                        ),
                    ]
                )
            else:
                expected_calls.append(
                    mock.call.check_call(["git", "checkout", "-b", branch])
                )
        assert mock_subprocess.mock_calls == expected_calls
        assert os.getcwd() == clone_dir

    # Clean up after exiting context
    assert not os.path.exists(clone_dir)
    assert os.getcwd() == initial_wd


@pytest.mark.parametrize(
    "branch, remote_branch_exists",
    [("master", True), ("other_test", True), ("other_test", False)],
)
def test_auto_config_updater_context_no_clone(
    branch, remote_branch_exists, tmpdir, mock_subprocess
):
    remote = "git_remote"
    working_dir = tmpdir.mkdir("testing")
    updater = config_utils.AutoConfigUpdater(
        "test_source",
        remote,
        branch=branch,
        working_dir=working_dir,
        do_clone=False,
    )
    updater._remote_branch_exists = mock.MagicMock(
        autospec=True, return_value=remote_branch_exists
    )
    initial_wd = os.getcwd()

    with updater:
        if branch == "master":
            expected_calls = []
        else:
            if remote_branch_exists:
                expected_calls = [
                    mock.call.check_call(["git", "fetch", "origin", branch]),
                    mock.call.check_call(
                        ["git", "checkout", "-b", branch, f"origin/{branch}"]
                    ),
                ]
            else:
                expected_calls = [
                    mock.call.check_call(["git", "checkout", "-b", branch])
                ]
        assert mock_subprocess.mock_calls == expected_calls
        assert os.getcwd() == working_dir

    # Existing directory not deleted
    assert os.path.exists(working_dir)
    assert os.getcwd() == initial_wd


@pytest.mark.parametrize("all_valid", [True, False])
@mock.patch("paasta_tools.config_utils.validate_auto_config_file", autospec=True)
def test_auto_config_updater_validate(mock_validate_file, all_valid, updater):
    mock_validate_file.side_effect = [True, all_valid, True]

    updater.write_configs("foo", "kubernetes-norcal-devc", {"a": 2})
    updater.write_configs("foo", "eks-pnw-devc", {"a": 2})
    assert updater.validate() == all_valid
    assert mock_validate_file.call_count == 2


def test_auto_config_updater_read_write(updater):
    service = "foo"
    conf_file = "something"
    data = {"key": "g_minor"}

    # Nothing yet
    assert updater.get_existing_configs(service, conf_file) == {}

    # Write something
    updater.write_configs(service, conf_file, data)
    assert updater.get_existing_configs(service, conf_file) == data
    assert len(updater.files_changed) == 1

    # Try writing to service directory that doesn't exist, should not change anything
    updater.write_configs("baz", conf_file, data)
    assert updater.get_existing_configs("baz", conf_file) == {}
    assert len(updater.files_changed) == 1


def test_auto_config_updater_read_write_with_comments(updater):
    service = "foo"
    conf_file = "something"
    data = {"key": "g_minor", "bar": "foo"}

    with open(f"{updater.working_dir}/{service}/{conf_file}.yaml", "x") as f:
        f.write(
            """# top comment
key: c_minor # inline comment
"""
        )

    existing_data = updater.get_existing_configs(service, conf_file)
    assert existing_data == {"key": "c_minor"}

    # Update existing config and add another key
    existing_data["key"] = "g_minor"
    existing_data["bar"] = "foo"

    updater.write_configs(service, conf_file, existing_data)
    assert updater.get_existing_configs(service, conf_file) == data
    assert len(updater.files_changed) == 1
    with open(f"{updater.working_dir}/{service}/{conf_file}.yaml", "r") as f:
        assert (
            f.read()
            == """# top comment
key: g_minor # inline comment
bar: foo
"""
        )


@mock.patch("paasta_tools.config_utils._push_to_remote", autospec=True)
@mock.patch("paasta_tools.config_utils._commit_files", autospec=True)
def test_auto_config_updater_commit_validate_fails(mock_push, mock_commit, updater):
    updater.files_changed = {"a", "b"}
    with mock.patch.object(
        updater, "validate", autospec=True, return_value=False
    ), pytest.raises(config_utils.ValidationError):
        updater.commit_to_remote()

    assert mock_commit.call_count == 0
    assert mock_push.call_count == 0


@pytest.mark.parametrize("did_commit", [True, False])
@mock.patch("paasta_tools.config_utils._commit_files", autospec=True)
@mock.patch("paasta_tools.config_utils._push_to_remote", autospec=True)
def test_auto_config_updater_commit(mock_push, mock_commit, did_commit, updater):
    updater.files_changed = {"a", "b"}
    mock_commit.return_value = did_commit
    with mock.patch.object(updater, "validate", autospec=True, return_value=True):
        updater.commit_to_remote()

    assert mock_commit.call_count == 1
    if did_commit:
        mock_push.assert_called_once_with(updater.branch)
    else:
        assert mock_push.call_count == 0


def test_auto_config_updater_merge_recommendations_limits(updater):
    service = "foo"
    conf_file = "notk8s-euwest-prod"
    limited_instance = "some_instance"
    noop_instance = "other_instance"
    autotune_data = {
        limited_instance: {"cpus": 0.1, "mem": 167, "disk": 256, "cpu_burst_add": 0.5}
    }
    user_data = {
        limited_instance: {
            "cmd": "ls",
            "autotune_limits": {
                "cpus": {"min": 1, "max": 2},
                "mem": {"min": 1024, "max": 2048},
                "disk": {"min": 512, "max": 1024},
            },
        },
        noop_instance: {"cmd": "ls"},
    }

    recs = {
        (service, conf_file): {
            limited_instance: {
                "mem": 1,
                "disk": 700,
                "cpus": 3,
            },
            noop_instance: {
                "cpus": 100,
                "mem": 10000,
                "disk": 2048,
            },
        }
    }

    with mock.patch.object(
        updater,
        "get_existing_configs",
        autospec=True,
        side_effect=[
            # first get the autotune data
            autotune_data,
            # then we get both the eks- and kuberentes- data
            user_data,
            # there could be data in both of these, but for a
            # simpler test, we just assume that we're looking
            # at something that's 100% on Yelp-managed k8s
            {},
        ],
    ):
        assert updater.merge_recommendations(recs) == {
            (service, conf_file): {
                limited_instance: {
                    "mem": 1024,  # use lower bound
                    "disk": 700,  # unchanged
                    "cpus": 2,  # use upper bound
                    "cpu_burst_add": 0.5,  # no updated rec or limit, leave alone
                },
                # this instances recommendations should be left untouched
                noop_instance: {
                    "cpus": 100,
                    "mem": 10000,
                    "disk": 2048,
                },
            }
        }



import argparse
import json
import logging
import sys
import time
from datetime import datetime
from datetime import timezone

from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAME
from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE
from paasta_tools.kubernetes_tools import get_namespaced_configmap
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import replace_namespaced_configmap

log = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description=f"Clean up expired entries from the {AUTOSCALING_OVERRIDES_CONFIGMAP_NAME} configmap"
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Print entries to be removed instead of removing them",
    )
    parser.add_argument(
        "-v", "--verbose", dest="verbose", action="store_true", default=False
    )
    args = parser.parse_args()
    return args


def setup_logging(verbose):
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(level=level)
    # silence kubernetes client metadata logs since debug ones are pretty noisy
    logging.getLogger("kubernetes.client.rest").setLevel(logging.ERROR)


def main():
    args = parse_args()
    setup_logging(args.verbose)

    kube_client = KubeClient()

    try:
        configmap = get_namespaced_configmap(
            name=AUTOSCALING_OVERRIDES_CONFIGMAP_NAME,
            namespace=AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE,
            kube_client=kube_client,
        )
    except ApiException as e:
        if e.status == 404:
            log.debug(
                f"ConfigMap {AUTOSCALING_OVERRIDES_CONFIGMAP_NAME} not found, nothing to clean up"
            )
            sys.exit(0)
        else:
            log.error(f"Error retrieving ConfigMap: {e}")
            sys.exit(1)

    if configmap is None:
        log.debug(
            f"ConfigMap {AUTOSCALING_OVERRIDES_CONFIGMAP_NAME} not found, nothing to clean up"
        )
        sys.exit(0)

    if not configmap.data:
        log.debug("ConfigMap has no data, nothing to clean up")
        sys.exit(0)

    current_timestamp = (
        time.time()
    )  # we'll fix the comparison time to the start of the script execution
    expired_entries = []  # used just for logging
    unexpired_data = {}  # used to replace the existing configmap data

    log.debug(f"Current timestamp: {current_timestamp}")
    for service_instance, override_json in configmap.data.items():
        log.debug(f"Processing entry for {service_instance}")

        try:
            override_data = json.loads(override_json)
        except json.JSONDecodeError:
            log.exception(
                f"Failed to parse JSON for {service_instance}, removing entry"
            )
            expired_entries.append(service_instance)
            continue

        if "expire_after" not in override_data:
            log.warning(
                f"Entry for {service_instance} missing expire_after field, removing entry"
            )
            expired_entries.append(service_instance)
            continue

        try:
            expire_after = float(override_data["expire_after"])
        except (ValueError, TypeError):
            log.exception(
                f"Failed to parse expire_after for {service_instance}, removing entry"
            )
            expired_entries.append(service_instance)
            continue

        if current_timestamp > expire_after:
            expire_datetime = datetime.fromtimestamp(expire_after, tz=timezone.utc)
            log.debug(
                f"Entry for {service_instance} expired at {expire_after} ({expire_datetime})"
            )
            expired_entries.append(service_instance)
        else:
            expire_datetime = datetime.fromtimestamp(expire_after, tz=timezone.utc)
            log.debug(
                f"Entry for {service_instance} expires at {expire_after} ({expire_datetime}), keeping"
            )
            unexpired_data[service_instance] = override_json

    if not expired_entries:
        log.info("No expired entries found")
        sys.exit(0)

    if args.dry_run:
        log.info(
            "Would have removed the following expired entries:\n "
            + "\n ".join(expired_entries)
        )
        sys.exit(0)

    try:
        configmap.data = (
            unexpired_data  # naughtily re-use the existing configmap object
        )
        replace_namespaced_configmap(  # NOTE: this is not a patch since you cannot currently remove entries in a patch
            name=AUTOSCALING_OVERRIDES_CONFIGMAP_NAME,
            namespace=AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE,
            body=configmap,
            kube_client=kube_client,
        )
        log.info(
            f"Successfully removed {len(expired_entries)} expired entries:\n "
            + "\n ".join(expired_entries)
        )
    except ApiException as e:
        log.error(f"Failed to update ConfigMap: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

from paasta_tools import deployment_utils
from paasta_tools.utils import DeploymentsJsonV2
from paasta_tools.utils import DeploymentVersion


@mock.patch("paasta_tools.deployment_utils.load_v2_deployments_json", autospec=True)
def test_get_currently_deployed_sha(
    mock_load_v2_deployments_json,
):
    mock_load_v2_deployments_json.return_value = DeploymentsJsonV2(
        service="fake-service",
        config_dict={
            "controls": {},
            "deployments": {"everything": {"git_sha": "abc", "docker_image": "foo"}},
        },
    )
    actual = deployment_utils.get_currently_deployed_sha(
        service="service", deploy_group="everything"
    )
    assert actual == "abc"


@mock.patch("paasta_tools.deployment_utils.load_v2_deployments_json", autospec=True)
def test_get_currently_deployed_version_no_image(
    mock_load_v2_deployments_json,
):
    mock_load_v2_deployments_json.return_value = DeploymentsJsonV2(
        service="fake-service",
        config_dict={
            "controls": {},
            "deployments": {"everything": {"git_sha": "abc", "docker_image": "foo"}},
        },
    )
    actual = deployment_utils.get_currently_deployed_version(
        service="service", deploy_group="everything"
    )
    assert actual == DeploymentVersion(sha="abc", image_version=None)


@mock.patch("paasta_tools.deployment_utils.load_v2_deployments_json", autospec=True)
def test_get_currently_deployed_version(
    mock_load_v2_deployments_json,
):
    mock_load_v2_deployments_json.return_value = DeploymentsJsonV2(
        service="fake-service",
        config_dict={
            "controls": {},
            "deployments": {
                "everything": {
                    "git_sha": "abc",
                    "docker_image": "foo",
                    "image_version": "extrastuff",
                }
            },
        },
    )
    actual = deployment_utils.get_currently_deployed_version(
        service="service", deploy_group="everything"
    )
    assert actual == DeploymentVersion(sha="abc", image_version="extrastuff")



from unittest import mock

from paasta_tools import check_services_replication_tools


def test_main_kubernetes():
    with mock.patch(
        "paasta_tools.check_services_replication_tools.check_services_replication",
        autospec=True,
    ) as mock_check_services_replication, mock.patch(
        "paasta_tools.check_services_replication_tools.parse_args", autospec=True
    ) as mock_parse_args, mock.patch(
        "paasta_tools.check_services_replication_tools.get_kubernetes_pods_and_nodes",
        autospec=True,
        return_value=([mock.Mock()], [mock.Mock()]),
    ), mock.patch(
        "paasta_tools.check_services_replication_tools.load_system_paasta_config",
        autospec=True,
    ) as mock_load_system_paasta_config, mock.patch(
        "paasta_tools.check_services_replication_tools.yelp_meteorite",
        autospec=True,
    ) as mock_yelp_meteorite, mock.patch(
        "paasta_tools.check_services_replication_tools.metrics_lib.system_timer",
        autospec=True,
    ) as mock_system_timer, mock.patch(
        "paasta_tools.check_services_replication_tools.sys.exit",
        autospec=True,
    ) as mock_sys_exit:
        mock_parse_args.return_value.under_replicated_crit_pct = 5
        mock_parse_args.return_value.min_count_critical = 1
        mock_parse_args.return_value.dry_run = False

        mock_check_services_replication.return_value = (6, 100)

        check_services_replication_tools.main(
            instance_type_class=None, check_service_replication=None, namespace="test"
        )
        assert mock_check_services_replication.called

        mock_yelp_meteorite.create_gauge.assert_called_once_with(
            "paasta.pct_services_under_replicated",
            {
                "paasta_cluster": mock_load_system_paasta_config.return_value.get_cluster.return_value,
                "scheduler": "kubernetes",
            },
        )
        mock_gauge = mock_yelp_meteorite.create_gauge.return_value
        mock_gauge.set.assert_called_once_with(6)

        mock_timer = mock_system_timer.return_value
        assert mock_timer.start.called
        mock_timer.stop.assert_called_once_with(tmp_dimensions={"result": 2})
        mock_sys_exit.assert_called_once_with(2)


def test_check_services_replication():
    soa_dir = "anw"
    instance_config = mock.Mock()
    instance_config.get_docker_image.return_value = True
    with mock.patch(
        "paasta_tools.check_services_replication_tools.list_services",
        autospec=True,
        return_value=["a"],
    ), mock.patch(
        "paasta_tools.check_kubernetes_services_replication.check_kubernetes_pod_replication",
        autospec=True,
    ) as mock_check_service_replication, mock.patch(
        "paasta_tools.check_services_replication_tools.PaastaServiceConfigLoader",
        autospec=True,
    ) as mock_paasta_service_config_loader, mock.patch(
        "paasta_tools.check_services_replication_tools.KubeClient", autospec=True
    ) as mock_kube_client:
        mock_kube_client.return_value = mock.Mock()
        mock_paasta_service_config_loader.return_value.instance_configs.return_value = [
            instance_config
        ]
        mock_client = mock.Mock()
        mock_client.list_tasks.return_value = []
        mock_replication_checker = mock.Mock()
        mock_pods = [mock.Mock(), mock.Mock()]
        mock_check_service_replication.return_value = True
        pods_by_service_instance = {
            instance_config.service: {
                instance_config.instance: mock_pods,
            }
        }

        (
            count_under_replicated,
            total,
        ) = check_services_replication_tools.check_services_replication(
            soa_dir=soa_dir,
            cluster="westeros-prod",
            service_instances=[],
            instance_type_class=None,
            check_service_replication=mock_check_service_replication,
            replication_checker=mock_replication_checker,
            pods_by_service_instance=pods_by_service_instance,
            dry_run=True,
        )
        mock_paasta_service_config_loader.assert_called_once_with(
            service="a", soa_dir=soa_dir
        )
        instance_config.get_docker_image.assert_called_once_with()
        mock_check_service_replication.assert_called_once_with(
            instance_config=instance_config,
            pods_by_service_instance=pods_by_service_instance,
            replication_checker=mock_replication_checker,
            dry_run=True,
        )
        assert count_under_replicated == 0
        assert total == 1



import sys
from unittest import mock

import pytest

from paasta_tools import spark_tools
from paasta_tools.spark_tools import auto_add_timeout_for_spark_job


def test_get_webui_url():
    with mock.patch("socket.getfqdn", return_value="1.2.3.4"):
        assert spark_tools.get_webui_url("1234") == "http://1.2.3.4:1234"


@pytest.mark.parametrize(
    "cmd,expected",
    [
        ("spark-shell", "spark-shell --conf spark.max.cores=100"),
        (
            "/venv/bin/pyspark test.py",
            "/venv/bin/pyspark --conf spark.max.cores=100 test.py",
        ),
        (
            "spark-submit script.py --other args",
            "spark-submit --conf spark.max.cores=100 script.py --other args",
        ),
        ("history-server", "history-server"),
    ],
)
def test_inject_spark_conf_str(cmd, expected):
    assert (
        spark_tools.inject_spark_conf_str(cmd, "--conf spark.max.cores=100") == expected
    )


@pytest.mark.parametrize(
    "spark_conf,expected",
    [
        (
            {
                "spark.kubernetes.executor.volumes.hostPath.nailsrv-123.mount.path": "/nail/srv",
                "spark.kubernetes.executor.volumes.hostPath.nailsrv-123.options.path": "/nail/srv",
                "spark.kubernetes.executor.volumes.hostPath.nailsrv-123.mount.readOnly": "true",
                "spark.kubernetes.executor.volumes.hostPath.123.mount.path": "/nail/123",
                "spark.kubernetes.executor.volumes.hostPath.123.options.path": "/nail/123",
                "spark.kubernetes.executor.volumes.hostPath.123.mount.readOnly": "false",
            },
            ["/nail/srv:/nail/srv:ro", "/nail/123:/nail/123:rw"],
        ),
        (
            {
                "spark.kubernetes.executor.volumes.hostPath.NAILsrv-123.mount.path": "/one/two",
                "spark.kubernetes.executor.volumes.hostPath.NAILsrv-123.options.path": "/one/two",
                "spark.kubernetes.executor.volumes.hostPath.NAILsrv-123.mount.readOnly": "true",
            },
            [""],
        ),
    ],
)
@mock.patch.object(sys, "exit")
def test_get_volumes_from_spark_k8s_configs(mock_sys, spark_conf, expected):
    result = spark_tools.get_volumes_from_spark_k8s_configs(spark_conf)
    if (
        "spark.kubernetes.executor.volumes.hostPath.NAILsrv-123.mount.path"
        in spark_conf
    ):
        mock_sys.assert_called_with(1)
    else:
        assert result == expected


@pytest.mark.parametrize(
    "spark_config,expected",
    [
        # Empty config
        ({}, {}),
        # Only UI port specified
        (
            {"spark.ui.port": "4040"},
            {
                "prometheus.io/port": "4040",
                "prometheus.io/path": "/metrics/prometheus",
            },
        ),
        # Only service and instance specified
        (
            {
                "spark.kubernetes.executor.annotation.paasta.yelp.com/service": "my-service",
                "spark.kubernetes.executor.annotation.paasta.yelp.com/instance": "my-instance",
            },
            {
                "paasta.yelp.com/service": "my-service",
                "paasta.yelp.com/instance": "my-instance",
            },
        ),
        # All annotations specified
        (
            {
                "spark.ui.port": "4040",
                "spark.kubernetes.executor.annotation.paasta.yelp.com/service": "my-service",
                "spark.kubernetes.executor.annotation.paasta.yelp.com/instance": "my-instance",
            },
            {
                "prometheus.io/port": "4040",
                "prometheus.io/path": "/metrics/prometheus",
                "paasta.yelp.com/service": "my-service",
                "paasta.yelp.com/instance": "my-instance",
            },
        ),
        # Missing service
        (
            {
                "spark.ui.port": "4040",
                "spark.kubernetes.executor.annotation.paasta.yelp.com/instance": "my-instance",
            },
            {
                "prometheus.io/port": "4040",
                "prometheus.io/path": "/metrics/prometheus",
            },
        ),
        # Missing instance
        (
            {
                "spark.ui.port": "4040",
                "spark.kubernetes.executor.annotation.paasta.yelp.com/service": "my-service",
            },
            {
                "prometheus.io/port": "4040",
                "prometheus.io/path": "/metrics/prometheus",
            },
        ),
    ],
)
def test_get_spark_driver_monitoring_annotations(spark_config, expected):
    result = spark_tools.get_spark_driver_monitoring_annotations(spark_config)
    assert result == expected


@pytest.mark.parametrize(
    argnames=[
        "cmd",
        "timeout_duration",
        "expected",
    ],
    argvalues=[
        pytest.param(
            "spark-submit abc.py",
            "4h",
            "timeout 4h spark-submit abc.py",
            id="No timeout",
        ),
        pytest.param(
            "timeout 2h spark-submit abc.py",
            "12h",
            "timeout 2h spark-submit abc.py",
            id="Timeout without options",
        ),
        pytest.param(
            "timeout -v 2h spark-submit abc.py",
            "12h",
            "timeout -v 2h spark-submit abc.py",
            id="Timeout with options",
        ),
        pytest.param(
            "timeout -v -s 1 2h spark-submit abc.py",
            "12h",
            "timeout -v -s 1 2h spark-submit abc.py",
            id="Timeout with multiple options",
        ),
        pytest.param(
            "timeout -k 10m --signal=SIGKILL 2h spark-submit abc.py",
            "12h",
            "timeout -k 10m --signal=SIGKILL 2h spark-submit abc.py",
            id="Timeout with double dash option",
        ),
    ],
)
def test_auto_add_timeout_for_spark_job(cmd, timeout_duration, expected):
    result = auto_add_timeout_for_spark_job(cmd, timeout_duration)

    assert result == expected



from unittest import mock

from paasta_tools.nrtsearchserviceeks_tools import (
    load_nrtsearchserviceeks_instance_config,
)


def test_load_nrtsearchserviceeks_instance_config():
    with mock.patch(
        "paasta_tools.nrtsearchserviceeks_tools.load_v2_deployments_json", autospec=True
    ) as mock_load_v2_deployments_json, mock.patch(
        "paasta_tools.nrtsearchserviceeks_tools.load_service_instance_config",
        autospec=True,
    ), mock.patch(
        "paasta_tools.nrtsearchserviceeks_tools.NrtsearchServiceEksDeploymentConfig",
        autospec=True,
    ) as mock_nrtsearchserviceeks_deployment_config:
        mock_config = {
            "port": None,
            "monitoring": {},
            "deploy": {},
            "data": {},
            "smartstack": {},
            "dependencies": {},
        }
        nrtsearchserviceeks_deployment_config = (
            load_nrtsearchserviceeks_instance_config(
                service="fake_nrtsearchservice_service",
                instance="fake_instance",
                cluster="fake_cluster",
                load_deployments=True,
                soa_dir="/foo/bar",
            )
        )
        mock_load_v2_deployments_json.assert_called_with(
            service="fake_nrtsearchservice_service", soa_dir="/foo/bar"
        )
        mock_nrtsearchserviceeks_deployment_config.assert_called_with(
            service="fake_nrtsearchservice_service",
            instance="fake_instance",
            cluster="fake_cluster",
            config_dict=mock_config,
            branch_dict=mock_load_v2_deployments_json.return_value.get_branch_dict(),
            soa_dir="/foo/bar",
        )

        assert (
            nrtsearchserviceeks_deployment_config
            == mock_nrtsearchserviceeks_deployment_config.return_value
        )
